# -*- coding: utf-8 -*-
import json
import os
import unittest
from collections import defaultdict
from pathlib import Path

import pytest
from PIL import Image
from pytest import raises

from kraken.containers import (BaselineLine, BaselineOCRRecord, BBoxLine,
                               BBoxOCRRecord, Segmentation)
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.models import load_any
from kraken.rpred import mm_rpred, rpred

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestBBoxRecords(unittest.TestCase):
    """
    Tests the bounding box OCR record.
    """
    def setUp(self):
        with open(resources / 'records.json', 'r') as fp:
            self.box_records = json.load(fp)
            self.ltr_record = self.box_records[0]

    def test_bbox_record_cuts(self):
        """
        Make sure that the cuts of a record are converted to absolute coordinates.
        """
        record = BBoxOCRRecord(**self.ltr_record)
        self.assertEqual(record.cuts, [[[1437, 119], [1437, 256], [1449, 256], [1449, 119]],
                                       [[1484, 119], [1484, 256], [1496, 256], [1496, 119]],
                                       [[1508, 119], [1508, 256], [1520, 256], [1520, 119]],
                                       [[1568, 119], [1568, 256], [1568, 256], [1568, 119]],
                                       [[1603, 119], [1603, 256], [1603, 256], [1603, 119]],
                                       [[1615, 119], [1615, 256], [1627, 256], [1627, 119]],
                                       [[1639, 119], [1639, 256], [1639, 256], [1639, 119]],
                                       [[1663, 119], [1663, 256], [1674, 256], [1674, 119]],
                                       [[1698, 119], [1698, 256], [1698, 256], [1698, 119]],
                                       [[1722, 119], [1722, 256], [1734, 256], [1734, 119]],
                                       [[1746, 119], [1746, 256], [1758, 256], [1758, 119]],
                                       [[1793, 119], [1793, 256], [1805, 256], [1805, 119]],
                                       [[1817, 119], [1817, 256], [1829, 256], [1829, 119]],
                                       [[1853, 119], [1853, 256], [1853, 256], [1853, 119]],
                                       [[1876, 119], [1876, 256], [1888, 256], [1888, 119]],
                                       [[1924, 119], [1924, 256], [1936, 256], [1936, 119]],
                                       [[1959, 119], [1959, 256], [1971, 256], [1971, 119]],
                                       [[2007, 119], [2007, 256], [2019, 256], [2019, 119]],
                                       [[2054, 119], [2054, 256], [2054, 256], [2054, 119]],
                                       [[2078, 119], [2078, 256], [2090, 256], [2090, 119]],
                                       [[2149, 119], [2149, 256], [2149, 256], [2149, 119]],
                                       [[2161, 119], [2161, 256], [2173, 256], [2173, 119]]])

    def test_bbox_record_redisplay(self):
        """
        Test that a display order record remains in display order when
        requesting a DO record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=True)
        self.assertEqual(record, record.display_order())

    def test_bbox_record_relogical(self):
        """
        Test that a logical order record remains in logical order when
        requesting a LO record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=False)
        self.assertEqual(record, record.logical_order())

    def test_bbox_record_display(self):
        """
        test display order conversion of record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=False)
        re_record = record.display_order()
        self.assertEqual(re_record.prediction, 'في معجزاته عليه السلام')
        self.assertEqual(re_record[:][1], ((1437, 119), (2173, 119), (2173, 256), (1437, 256)))
        self.assertAlmostEqual(re_record[2:8][2], 0.9554762, places=4)

    def test_bbox_record_logical(self):
        """
        Test logical order conversion of record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=True)
        re_record = record.logical_order()
        self.assertEqual(re_record.prediction, 'في معجزاته عليه السلام')
        self.assertEqual(re_record[:][1], ((1437, 119), (2173, 119), (2173, 256), (1437, 256)))
        self.assertAlmostEqual(re_record[2:8][2], 0.9554762, places=4)

    def test_bbox_record_slicing(self):
        """
        Tests simple slicing/aggregation of elements in record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=True)
        pred, cut, conf = record[1:8]
        self.assertEqual(pred, 'السلا ه')
        self.assertEqual(cut, ((1484, 119), (1674, 119), (1674, 256), (1484, 256)))
        self.assertAlmostEqual(conf, 0.9259478, places=4)

    def test_bbox_record_slicing(self):
        """
        Tests complex slicing/aggregation of elements in record.
        """
        record = BBoxOCRRecord(**self.ltr_record, display_order=True)
        pred, cut, conf = record[1:5:2]
        self.assertEqual(pred, 'اس')
        self.assertEqual(cut, ((1484, 119), (1568, 119), (1568, 256), (1484, 256)))
        self.assertAlmostEqual(conf, 0.74411, places=4)


class TestBaselineRecords(unittest.TestCase):
    """
    Tests the baseline OCR record.
    """
    def setUp(self):
        with open(resources / 'bl_records.json', 'r') as fp:
            self.bl_records = json.load(fp)
            self.ltr_record = self.bl_records['lines'][15]

    def test_baseline_record_cuts(self):
        """
        Make sure that the cuts of a record are converted to absolute coordinates.
        """
        record = BaselineOCRRecord(**self.ltr_record)

    def test_baseline_record_redisplay(self):
        """
        Test that a display order record remains in display order when
        requesting a DO record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=True)
        self.assertEqual(record, record.display_order())

    def test_baseline_record_relogical(self):
        """
        Test that a logical order record remains in logical order when
        requesting a LO record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=False)
        self.assertEqual(record, record.logical_order())

    def test_baseline_record_display(self):
        """
        test display order conversion of record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=False)
        re_record = record.display_order()
        self.assertEqual(re_record.prediction, '.هنيدو')
        self.assertEqual(re_record[:][1], ([1370, 1382], [1370, 1424], [1448, 1426], [1451, 1388]))
        self.assertAlmostEqual(re_record[2:4][2], 0.9998551, places=4)

    def test_baseline_record_logical(self):
        """
        Test logical order conversion of record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=True)
        re_record = record.logical_order()
        self.assertEqual(re_record.prediction, '.هنيدو')
        self.assertEqual(re_record[:][1], ([1370, 1382], [1370, 1424], [1448, 1426], [1451, 1388]))
        self.assertAlmostEqual(re_record[2:3][2], 0.99996733, places=4)

    def test_baseline_record_slicing(self):
        """
        Tests simple slicing/aggregation of elements in record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=True)
        pred, cut, conf = record[2:5]
        self.assertEqual(pred, 'ينه')
        self.assertEqual(cut, ([1385, 1375], [1385, 1427], [1411, 1433], [1411, 1378]))
        self.assertAlmostEqual(conf, 0.99957436, places=4)

    def test_baseline_complex_record_slicing(self):
        """
        Tests complex slicing/aggregation of elements in record.
        """
        record = BaselineOCRRecord(**self.ltr_record, display_order=True)
        pred, cut, conf = record[1:5:2]
        self.assertEqual(pred, 'دن')
        self.assertEqual(cut, ([1396, 1375], [1396, 1430], [1426, 1437], [1429, 1381]))
        self.assertAlmostEqual(conf, 0.999982893, places=4)


class TestRecognition(unittest.TestCase):

    """
    Tests of the recognition facility and associated routines.
    """
    def setUp(self):
        self.im = Image.open(resources / 'bw.png')
        self.overfit_line = Image.open(resources / '000236.png')
        self.model = load_any(resources / 'overfit.mlmodel')
        self.invalid_box_seg = Segmentation(type='bbox',
                                            imagename = resources / 'bw.png',
                                            lines=[BBoxLine(id='foo',
                                                            bbox=[-1, -1, 10000, 10000])],
                                            text_direction='horizontal-lr',
                                            script_detection=False
                                           )
        self.invalid_bl_seg = Segmentation(type='baselines',
                                           imagename = resources / 'bw.png',
                                           lines=[BaselineLine(id='bar',
                                                               tags={'type': 'default'},
                                                               baseline=[[0,0], [10000, 0]],
                                                               boundary=[[-1, -1], [-1, 10000], [10000, 10000], [10000, -1]])],
                                           text_direction='horizontal-lr',
                                           script_detection=False
                                          )

        self.simple_box_seg = Segmentation(type='bbox',
                                           imagename = resources / 'bw.png',
                                           lines=[BBoxLine(id='foo',
                                                           bbox=[0, 0, 2544, 156])],
                                           text_direction='horizontal-lr',
                                           script_detection=False
                                          )
        self.simple_bl_seg = Segmentation(type='baselines',
                                          imagename = resources / 'bw.png',
                                          lines=[BaselineLine(id='foo',
                                                              baseline=[[0, 10], [2543, 10]],
                                                              boundary=[[0, 0], [2543, 0], [2543, 155], [0, 155]])],
                                           text_direction='horizontal-lr',
                                           script_detection=False
                                         )

        self.tagged_box_seg = Segmentation(type='bbox',
                                           imagename = resources / 'bw.png',
                                           lines=[BBoxLine(id='foo',
                                                           bbox=[0, 0, 2544, 156],
                                                           tags={'type': 'foobar'}),
                                                  BBoxLine(id='bar',
                                                           bbox=[0, 0, 2544, 156],
                                                           tags={'type': 'default'})
                                                 ],
                                           text_direction='horizontal-lr',
                                           script_detection=True
                                          )
        self.tagged_bl_seg = Segmentation(type='baselines',
                                          imagename = resources / 'bw.png',
                                          lines=[BaselineLine(id='foo',
                                                              baseline=[[0, 10], [2543, 10]],
                                                              boundary=[[0, 0], [2543, 0], [2543, 155], [0, 155]],
                                                              tags={'type': 'foobar'}),
                                                 BaselineLine(id='bar',
                                                              baseline=[[0, 10], [2543, 10]],
                                                              boundary=[[0, 0], [2543, 0], [2543, 155], [0, 155]],
                                                              tags={'type': 'default'}),
                                                ],
                                           text_direction='horizontal-lr',
                                           script_detection=True
                                          )


    def test_rpred_bbox_outbounds(self):
        """
        Tests correct handling of invalid bbox line coordinates.
        """
        with raises(KrakenInputException):
            pred = rpred(self.model, self.im, self.invalid_box_seg, True)
            next(pred)

    @pytest.mark.xfail
    def test_rpred_bl_outbounds(self):
        """
        Tests correct handling of invalid baseline coordinates.
        """
        with raises(KrakenInputException):
            pred = rpred(self.model, self.im, self.invalid_bl_seg, True)
            next(pred)

    def test_simple_bbox_rpred(self):
        """
        Tests simple recognition without tags.
        """
        pred = rpred(self.model, self.overfit_line, self.simple_box_seg, True)
        record = next(pred)
        self.assertEqual(record.prediction, 'ܡ ܘܡ ܗ ܡܕܐ ܐ ܐܐ ܡ ܗܗܐܐܐܕ')

    def test_simple_bl_rpred(self):
        """
        Tests simple recognition without tags.
        """
        pred = rpred(self.model, self.overfit_line, self.simple_bl_seg, True)
        record = next(pred)
        self.assertEqual(record.prediction, '.ܗ ܣܗܐ  ܕ ܣ   ܗ ܕܗܗ ܟܕܗܣ    ܠ  ܐ .ܣܕܐܣ. ܗ ')

    def test_mm_rpred_bbox_missing_tags(self):
        """
        Test that mm_rpred fails when tags are missing
        """
        with raises(ValueError):
            pred = mm_rpred({('type', 'default'): self.model},
                            self.overfit_line,
                            self.simple_box_seg,
                            True)

    def test_mm_rpred_bl_missing_tags(self):
        """
        Test that mm_rpred fails when tags are missing
        """
        with raises(ValueError):
            pred = mm_rpred({('type', 'default'): self.model},
                            self.overfit_line,
                            self.simple_bl_seg,
                            True)

    def test_mm_rpred_bbox_ignore_tags(self):
        """
        Tests mm_rpred recognition with ignore tags.
        """
        pred = mm_rpred({('type', 'default'): self.model},
                        self.overfit_line,
                        self.tagged_box_seg,
                        True,
                        tags_ignore=[('type', 'foobar')])
        record = next(pred)
        self.assertEqual(record.prediction, '')
        record = next(pred)
        self.assertEqual(record.prediction, 'ܡ ܘܡ ܗ ܡܕܐ ܐ ܐܐ ܡ ܗܗܐܐܐܕ')

    def test_mm_rpred_bbox_default_tags(self):
        """
        Tests recognition with default tag.
        """
        pred = mm_rpred(defaultdict(lambda: self.model),
                        self.overfit_line,
                        self.tagged_box_seg,
                        True)
        record = next(pred)
        self.assertEqual(record.prediction, 'ܡ ܘܡ ܗ ܡܕܐ ܐ ܐܐ ܡ ܗܗܐܐܐܕ')
        record = next(pred)
        self.assertEqual(record.prediction, 'ܡ ܘܡ ܗ ܡܕܐ ܐ ܐܐ ܡ ܗܗܐܐܐܕ')

    def test_mm_rpred_bl_ignore_tags(self):
        """
        Tests baseline recognition with ignore tags.
        """
        pred = mm_rpred({('type', 'default'): self.model},
                        self.overfit_line,
                        self.tagged_bl_seg,
                        True,
                        tags_ignore=[('type', 'foobar')])
        record = next(pred)
        self.assertEqual(record.prediction, '')
        record = next(pred)
        self.assertEqual(record.prediction, '.ܗ ܣܗܐ  ܕ ܣ   ܗ ܕܗܗ ܟܕܗܣ    ܠ  ܐ .ܣܕܐܣ. ܗ ')

    def test_mm_rpred_bl_default_tags(self):
        """
        Tests baseline recognition with default tag.
        """
        pred = mm_rpred(defaultdict(lambda: self.model),
                        self.overfit_line,
                        self.tagged_bl_seg,
                        True)
        record = next(pred)
        self.assertEqual(record.prediction, '.ܗ ܣܗܐ  ܕ ܣ   ܗ ܕܗܗ ܟܕܗܣ    ܠ  ܐ .ܣܕܐܣ. ܗ ')
        record = next(pred)
        self.assertEqual(record.prediction, '.ܗ ܣܗܐ  ܕ ܣ   ܗ ܕܗܗ ܟܕܗܣ    ܠ  ܐ .ܣܕܐܣ. ܗ ')

    def test_mm_rpred_bl_nobidi(self):
        """
        Tests baseline recognition without bidi reordering.
        """
        pred = mm_rpred(defaultdict(lambda: self.model),
                        self.overfit_line,
                        self.simple_bl_seg,
                        bidi_reordering=False)
        record = next(pred)
        self.assertEqual(record.prediction, 'ܕܗ .ܣܐܗܗ.ܐ ܗܣ ܕ   ܗܣ ܗ.ܗܝܣܗ ܣ ܗܢ ܪܗܗܕ ܐ   ܗܠ')

    def test_mm_rpred_bbox_nobidi(self):
        """
        Tests bbox recognition without bidi reordering.
        """
        pred = mm_rpred(defaultdict(lambda: self.model),
                        self.overfit_line,
                        self.simple_box_seg,
                        bidi_reordering=False)
        record = next(pred)
        self.assertEqual(record.prediction,  'ܕܗܣܐܕ ܪܝ .ܡܡ ܐܠܠ ܗܠ ܐܘܗ ܟܘܗܢ ܡܡ ܐܠ')



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.dataset
~~~~~~~~~~~~~~~~~~~~

Command line driver for dataset compilation
"""
import click

from .util import _validate_manifests


@click.command('compile')
@click.pass_context
@click.option('-o', '--output', show_default=True, type=click.Path(), default='dataset.arrow', help='Output dataset file')
@click.option('--workers', show_default=True, default=1, help='Number of parallel workers for text line extraction.')
@click.option('-f', '--format-type', type=click.Choice(['path', 'xml', 'alto', 'page']), default='xml', show_default=True,
              help='Sets the training data format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both baselines and a '
              'link to source images. In `path` mode arguments are image files '
              'sharing a prefix up to the last extension with JSON `.path` files '
              'containing the baseline information.')
@click.option('-F', '--files', show_default=True, default=None, multiple=True,
              callback=_validate_manifests, type=click.File(mode='r', lazy=True),
              help='File(s) with additional paths to training data.')
@click.option('--random-split', type=float, nargs=3, default=None, show_default=True,
              help='Creates a fixed random split of the input data with the '
              'proportions (train, validation, test). Overrides the save split option.')
@click.option('--force-type', type=click.Choice(['bbox', 'baseline']), default=None, show_default=True,
              help='Forces the dataset type to a specific value. Can be used to '
                   '"convert" a line strip-type collection to a baseline-style '
                   'dataset, e.g. to disable centerline normalization.')
@click.option('--save-splits/--ignore-splits', show_default=True, default=True,
              help='Whether to serialize explicit splits contained in XML '
                   'files. Is ignored in `path` mode.')
@click.option('--skip-empty-lines/--keep-empty-lines', show_default=True, default=True,
              help='Whether to keep or skip empty text lines. Text-less '
                   'datasets are useful for unsupervised pretraining but '
                   'loading datasets with many empty lines for recognition '
                   'training is inefficient.')
@click.option('--recordbatch-size', show_default=True, default=100,
              help='Minimum number of records per RecordBatch written to the '
                   'output file. Larger batches require more transient memory '
                   'but slightly improve reading performance.')
@click.option('--legacy-polygons', show_default=True, default=False, is_flag=True,
                help='Use the old polygon extractor.')
@click.argument('ground_truth', nargs=-1, type=click.Path(exists=True, dir_okay=False))
def compile(ctx, output, workers, format_type, files, random_split, force_type,
            save_splits, skip_empty_lines, recordbatch_size, ground_truth, legacy_polygons):
    """
    Precompiles a binary dataset from a collection of XML files.
    """
    from kraken.lib.progress import KrakenProgressBar

    from .util import message

    ground_truth = list(ground_truth)

    if files:
        ground_truth.extend(files)

    if not ground_truth:
        raise click.UsageError('No training data was provided to the compile command. Use the `ground_truth` argument.')

    from kraken.lib import arrow_dataset

    force_type = {'bbox': 'kraken_recognition_bbox',
                  'baseline': 'kraken_recognition_baseline',
                  None: None}[force_type]

    with KrakenProgressBar() as progress:
        extract_task = progress.add_task('Extracting lines', total=0, start=False, visible=True if not ctx.meta['verbose'] else False)

        def _update_bar(advance, total):
            if not progress.tasks[0].started:
                progress.start_task(extract_task)
            progress.update(extract_task, total=total, advance=advance)

        arrow_dataset.build_binary_dataset(ground_truth,
                                           output,
                                           format_type,
                                           workers,
                                           save_splits,
                                           random_split,
                                           force_type,
                                           recordbatch_size,
                                           skip_empty_lines,
                                           _update_bar,
                                           legacy_polygons=legacy_polygons)

    message(f'Output file written to {output}')



# -*- coding: utf-8 -*-
import os
import tempfile
import unittest
from pathlib import Path

import click
import numpy as np
from click.testing import CliRunner
from PIL import Image
from pytest import raises

from kraken.kraken import cli

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestCLI(unittest.TestCase):
    """
    Testing the kraken CLI
    """

    def setUp(self):
        self.temp = tempfile.NamedTemporaryFile(delete=False)
        self.runner = CliRunner()
        self.color_img = resources / 'input.tif'
        self.bw_img = resources / 'bw.png'

    def tearDown(self):
        self.temp.close()
        os.unlink(self.temp.name)

    def test_binarize_color(self):
        """
        Tests binarization of color images.
        """
        with tempfile.NamedTemporaryFile() as fp:
            result = self.runner.invoke(cli, ['-i', self.color_img, fp.name, 'binarize'])
            self.assertEqual(result.exit_code, 0)
            self.assertEqual(tuple(map(lambda x: x[1], Image.open(fp).getcolors())), (0, 255))

    def test_binarize_bw(self):
        """
        Tests binarization of b/w images.
        """
        with tempfile.NamedTemporaryFile() as fp:
            result = self.runner.invoke(cli, ['-i', self.bw_img, fp.name, 'binarize'])
            self.assertEqual(result.exit_code, 0)
            bw = np.array(Image.open(self.bw_img))
            new = np.array(Image.open(fp.name))
            self.assertTrue(np.all(bw == new))

    def test_segment_color(self):
        """
        Tests that segmentation is aborted when given color image.
        """
        with tempfile.NamedTemporaryFile() as fp:
            result = self.runner.invoke(cli, ['-r', '-i', self.color_img, fp.name, 'segment'])
            self.assertEqual(result.exit_code, 1)

    def test_segment_color_noraise(self):
        """
        Tests that segmentation does not return 1 when given color image in noraise mode.
        """
        with tempfile.NamedTemporaryFile() as fp:
            result = self.runner.invoke(cli, ['-i', self.color_img, fp.name, 'segment'])
            self.assertEqual(result.exit_code, 0)



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.util
~~~~~~~~~~~~~~~~~~~~

Command line driver helpers
"""
import glob
import logging
import os
from typing import List, Optional, Tuple

import click

logging.captureWarnings(True)
logger = logging.getLogger('kraken')


def _validate_manifests(ctx, param, value):
    images = []
    for manifest in value:
        try:
            for entry in manifest.readlines():
                im_p = entry.rstrip('\r\n')
                if os.path.isfile(im_p):
                    images.append(im_p)
                else:
                    logger.warning('Invalid entry "{}" in {}'.format(im_p, manifest.name))
        except UnicodeDecodeError:
            raise click.BadOptionUsage(param,
                                       f'File {manifest.name} is not a text file. Please '
                                        'ensure that the argument to `-t`/`-e` is a manifest '
                                        'file containing paths to training data (one per '
                                        'line).',
                                       ctx=ctx)
    return images


def _expand_gt(ctx, param, value):
    images = []
    for expression in value:
        images.extend([x for x in glob.iglob(expression, recursive=True) if os.path.isfile(x)])
    return images


def message(msg, **styles):
    if logger.getEffectiveLevel() >= 30:
        click.secho(msg, **styles)


def to_ptl_device(device: str) -> Tuple[str, Optional[List[int]]]:
    if device in ['cpu', 'mps']:
        return device, 'auto'
    elif any([device.startswith(x) for x in ['tpu', 'cuda', 'hpu', 'ipu']]):
        dev, idx = device.split(':')
        if dev == 'cuda':
            dev = 'gpu'
        return dev, [int(idx)]
    raise Exception(f'Invalid device {device} specified')



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from typing import Sequence, Tuple, Union

import numpy as np
import torch


def positive_integers_with_sum(n, total):
    ls = [0]
    rv = []
    while len(ls) < n:
        c = random.randint(0, total)
        ls.append(c)
    ls = sorted(ls)
    ls.append(total)
    for i in range(1, len(ls)):
        rv.append(ls[i] - ls[i-1])
    return rv


def compute_masks(mask_prob: int,
                  mask_width: int,
                  num_neg_samples: int,
                  seq_lens: Union[torch.Tensor, Sequence[int]]):
    """
    Samples num_mask non-overlapping random masks of length mask_width in
    sequence of length seq_len.

    Args:
        mask_prob: Probability of each individual token being chosen as start
                   of a masked sequence. Overall number of masks num_masks is
                   mask_prob * sum(seq_lens) / mask_width.
        mask_width: width of each mask
        num_neg_samples: Number of samples from unmasked sequence parts (gets
                         multiplied by num_mask)
        seq_lens: sequence lengths

    Returns:
        An index array containing 1 for masked bits, 2 for negative samples,
        the number of masks, and the actual number of negative samples.
    """
    mask_samples = np.zeros(sum(seq_lens))
    num_masks = int(mask_prob * sum(seq_lens.numpy()) // mask_width)
    num_neg_samples = num_masks * num_neg_samples
    num_masks += num_neg_samples

    indices = [x+mask_width for x in positive_integers_with_sum(num_masks, sum(seq_lens)-num_masks*mask_width)]
    start = 0
    mask_slices = []
    for i in indices:
        i_start = random.randint(start, i+start-mask_width)
        mask_slices.append(slice(i_start, i_start+mask_width))
        start += i

    neg_idx = random.sample(range(len(mask_slices)), num_neg_samples)
    neg_slices = [mask_slices.pop(idx) for idx in sorted(neg_idx, reverse=True)]

    mask_samples[np.r_[tuple(mask_slices)]] = 1
    mask_samples[np.r_[tuple(neg_slices)]] = 2

    return mask_samples, num_masks - num_neg_samples, num_neg_samples


def buffered_arange(max):
    if not hasattr(buffered_arange, "buf"):
        buffered_arange.buf = torch.LongTensor()
    if max > buffered_arange.buf.numel():
        buffered_arange.buf.resize_(max)
        torch.arange(max, out=buffered_arange.buf)
    return buffered_arange.buf[:max]


def sample_negatives(y, num_samples, num_neg_samples: int):
    B, W, C = y.shape
    y = y.view(-1, C)  # BTC => (BxT)C

    with torch.no_grad():
        tszs = (buffered_arange(num_samples).unsqueeze(-1).expand(-1, num_neg_samples).flatten())

        neg_idxs = torch.randint(low=0, high=W - 1, size=(B, num_neg_samples * num_samples))
        neg_idxs[neg_idxs >= tszs] += 1

    for i in range(1, B):
        neg_idxs[i] += i * W

    negs = y[neg_idxs.view(-1)]
    negs = negs.view(B, num_samples, num_neg_samples, C).permute(2, 0, 1, 3)  # to NxBxTxC
    return negs


def compute_mask_indices(shape: Tuple[int, int], mask_prob: float, mask_length: int = 4, mask_min_space: int = 2) -> np.ndarray:
    """
    Computes random mask spans for a given shape

    Args:
        shape: the the shape for which to compute masks.
            should be of size 2 where first element is batch size and 2nd is timesteps
        mask_prob: probability for each token to be chosen as start of the span to be masked. this will be multiplied by
            number of timesteps divided by length of mask span to mask approximately this percentage of all elements.
            however due to overlaps, the actual number will be smaller.
    """

    bsz, all_sz = shape
    mask = np.full((bsz, all_sz), False)

    all_num_mask = int(mask_prob * all_sz / float(mask_length) + np.random.rand())

    mask_idcs = []
    for i in range(bsz):
        # import ipdb; ipdb.set_trace()
        sz = all_sz
        num_mask = all_num_mask

        lengths = np.full(num_mask, mask_length)

        if sum(lengths) == 0:
            lengths[0] = min(mask_length, sz - 1)

        mask_idc = []

        def arrange(s, e, length, keep_length):
            span_start = np.random.randint(s, e - length)
            mask_idc.extend(span_start + i for i in range(length))

            new_parts = []
            if span_start - s - mask_min_space >= keep_length:
                new_parts.append((s, span_start - mask_min_space + 1))
            if e - span_start - keep_length - mask_min_space > keep_length:
                new_parts.append((span_start + length + mask_min_space, e))
            return new_parts

        parts = [(0, sz)]
        min_length = min(lengths)
        for length in sorted(lengths, reverse=True):
            lens = np.fromiter(
                (e - s if e - s >= length + mask_min_space else 0 for s, e in parts),
                int,
            )
            l_sum = np.sum(lens)
            if l_sum == 0:
                break
            probs = lens / np.sum(lens)
            c = np.random.choice(len(parts), p=probs)
            s, e = parts.pop(c)
            parts.extend(arrange(s, e, length, min_length))
        mask_idc = np.asarray(mask_idc)

        mask_idcs.append(np.unique(mask_idc[mask_idc < sz]))

    # make sure all masks are the same length in the batch by removing masks
    # if they are greater than the min length mask
    min_len = min([len(m) for m in mask_idcs])

    for i, mask_idc in enumerate(mask_idcs):
        if len(mask_idc) > min_len:
            mask_idc = np.random.choice(mask_idc, min_len, replace=False)
        assert len(mask_idc) == min_len
        mask[i, mask_idc] = True

    return mask



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Pytorch-lightning modules for recognition model pretraining.

Pretraining is based on an image inpainting surrogate task that aims to
reconstruct randomly sampled masked patches from the initial convolutional
feature maps that have been replaced with a learnable embedding. The model is
trained with a contrastive loss where negative samples are randomly generated
from the unmasked parts of the sequence.

Apart from an improved sampling method the implementation is mostly a faithful
adaptation of:

Vogler, Nikolai, et al. "Lacuna Reconstruction: Self-supervised Pre-training
for Low-Resource Historical Document Transcription." arXiv preprint
arXiv:2112.08692 (2021).
"""
import logging
import math
import re
from itertools import chain
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Union

import numpy as np
import lightning as L
import torch
import torch.nn.functional as F
from lightning.pytorch.callbacks import EarlyStopping
from lightning.pytorch.utilities.memory import (garbage_collection_cuda,
                                                is_oom_error)
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader, Subset, random_split

from kraken.containers import Segmentation
from kraken.lib import default_specs, layers, vgsl
from kraken.lib.codec import PytorchCodec
from kraken.lib.dataset import (ArrowIPCRecognitionDataset, GroundTruthDataset,
                                ImageInputTransforms, PolygonGTDataset,
                                collate_sequences)
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.pretrain.layers import Wav2Vec2Mask
from kraken.lib.train import _configure_optimizer_and_lr_scheduler
from kraken.lib.util import parse_gt_path
from kraken.lib.xml import XMLPage

if TYPE_CHECKING:
    from os import PathLike

logger = logging.getLogger(__name__)


def _star_fun(fun, kwargs):
    try:
        return fun(**kwargs)
    except FileNotFoundError as e:
        logger.warning(f'{e.strerror}: {e.filename}. Skipping.')
    except KrakenInputException as e:
        logger.warning(str(e))
    return None


class PretrainDataModule(L.LightningDataModule):
    def __init__(self,
                 training_data: Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]] = None,
                 evaluation_data: Optional[Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]]] = None,
                 partition: Optional[float] = 0.9,
                 binary_dataset_split: bool = False,
                 batch_size: int = 4,
                 height: int = 48,
                 width: int = 0,
                 channels: int = 1,
                 num_workers: int = 1,
                 force_binarization: bool = False,
                 format_type: str = 'path',
                 pad: int = 16,
                 augment: bool = default_specs.RECOGNITION_PRETRAIN_HYPER_PARAMS['augment'],
                 legacy_polygons: bool = False):
        """
        A LightningDataModule encapsulating text-less training data for
        unsupervised recognition model pretraining.

        Args:
            training_data:
            evaluation_data:
            partition:
            binary_dataset_split:
            batch_size:
            num_workers:
            force_binarization:
            format_type:
            augment:
        """
        super().__init__()
        self.save_hyperparameters()

        self.legacy_polygons = legacy_polygons

        DatasetClass = GroundTruthDataset
        valid_norm = True
        if format_type in ['xml', 'page', 'alto']:
            logger.info(f'Parsing {len(training_data)} XML files for training data')
            training_data = [{'page': XMLPage(file, format_type).to_container()} for file in training_data]
            if evaluation_data:
                logger.info(f'Parsing {len(evaluation_data)} XML files for validation data')
                evaluation_data = [{'page': XMLPage(file, format_type).to_container()} for file in evaluation_data]
            if binary_dataset_split:
                logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                binary_dataset_split = False
            DatasetClass = partial(PolygonGTDataset, legacy_polygons=legacy_polygons)
            valid_norm = False
        elif format_type == 'binary':
            DatasetClass = ArrowIPCRecognitionDataset
            valid_norm = False
            logger.info(f'Got {len(training_data)} binary dataset files for training data')
            training_data = [{'file': file} for file in training_data]
            if evaluation_data:
                logger.info(f'Got {len(evaluation_data)} binary dataset files for validation data')
                evaluation_data = [{'file': file} for file in evaluation_data]
        elif format_type == 'path':
            if force_binarization:
                logger.warning('Forced binarization enabled in `path` mode. Will be ignored.')
                force_binarization = False
            if binary_dataset_split:
                logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                binary_dataset_split = False
            logger.info(f'Got {len(training_data)} line strip images for training data')
            training_data = [{'line': parse_gt_path(im)} for im in training_data]
            if evaluation_data:
                logger.info(f'Got {len(evaluation_data)} line strip images for validation data')
                evaluation_data = [{'line': parse_gt_path(im)} for im in evaluation_data]
            valid_norm = True
        # format_type is None. Determine training type from length of training data entry
        elif not format_type:
            if training_data[0].type == 'baselines':
                DatasetClass = partial(PolygonGTDataset, legacy_polygons=legacy_polygons)
                valid_norm = False
            else:
                if force_binarization:
                    logger.warning('Forced binarization enabled with box lines. Will be ignored.')
                    force_binarization = False
                if binary_dataset_split:
                    logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                    binary_dataset_split = False
            samples = []
            for sample in training_data:
                if isinstance(sample, Segmentation):
                    samples.append({'page': sample})
                else:
                    samples.append({'line': sample})
            training_data = samples
            if evaluation_data:
                samples = []
                for sample in evaluation_data:
                    if isinstance(sample, Segmentation):
                        samples.append({'page': sample})
                    else:
                        samples.append({'line': sample})
                evaluation_data = samples

        else:
            raise ValueError(f'format_type {format_type} not in [alto, page, xml, path, binary].')

        self.transforms = ImageInputTransforms(batch_size,
                                               height,
                                               width,
                                               channels,
                                               (pad, 0),
                                               valid_norm,
                                               force_binarization)

        if evaluation_data:
            train_set = self._build_dataset(DatasetClass, training_data)
            self.train_set = Subset(train_set, range(len(train_set)))
            val_set = self._build_dataset(DatasetClass, evaluation_data)
            self.val_set = Subset(val_set, range(len(val_set)))
        elif binary_dataset_split:
            train_set = self._build_dataset(DatasetClass, training_data, split_filter='train')
            self.train_set = Subset(train_set, range(len(train_set)))
            val_set = self._build_dataset(DatasetClass, training_data, split_filter='validation')
            self.val_set = Subset(val_set, range(len(val_set)))
            logger.info(f'Found {len(self.train_set)} (train) / {len(self.val_set)} (val) samples in pre-encoded dataset')
        else:
            train_set = self._build_dataset(DatasetClass, training_data)
            train_len = int(len(train_set)*partition)
            val_len = len(train_set) - train_len
            logger.info(f'No explicit validation data provided. Splitting off '
                        f'{val_len} (of {len(train_set)}) samples to validation '
                        'set. (Will disable alphabet mismatch detection.)')
            self.train_set, self.val_set = random_split(train_set, (train_len, val_len))

        if format_type == 'binary':
            legacy_train_status = self.train_set.dataset.legacy_polygons_status
            if self.val_set.dataset.legacy_polygons_status != legacy_train_status:
                logger.warning('Train and validation set have different legacy '
                               f'polygon status: {legacy_train_status} and '
                               f'{self.val_set.dataset.legacy_polygons_status}. Train set '
                               'status prevails.')
            if legacy_train_status == "mixed":
                logger.warning('Mixed legacy polygon status in training dataset. Consider recompilation.')
                legacy_train_status = False
            if legacy_polygons != legacy_train_status:
                logger.warning(f'Setting dataset legacy polygon status to {legacy_train_status} based on training set.')
                self.legacy_polygons = legacy_train_status

        if len(self.train_set) == 0 or len(self.val_set) == 0:
            raise ValueError('No valid training data was provided to the train '
                             'command. Please add valid XML, line, or binary data.')

        logger.info(f'Training set {len(self.train_set)} lines, validation set '
                    f'{len(self.val_set)} lines')

    def _build_dataset(self,
                       DatasetClass,
                       training_data,
                       **kwargs):
        dataset = DatasetClass(im_transforms=self.transforms,
                               augmentation=self.hparams.augment,
                               skip_empty_lines=False,
                               **kwargs)

        for sample in training_data:
            try:
                dataset.add(**sample)
            except KrakenInputException as e:
                logger.warning(str(e))

        return dataset

    def train_dataloader(self):
        return DataLoader(self.train_set,
                          collate_fn=collate_sequences,
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_workers,
                          pin_memory=True)

    def val_dataloader(self):
        return DataLoader(self.val_set,
                          collate_fn=collate_sequences,
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_workers,
                          pin_memory=True)

    def setup(self, stage: Optional[str] = None):
        self.train_set.dataset.no_encode()
        self.val_set.dataset.no_encode()


class RecognitionPretrainModel(L.LightningModule):
    def __init__(self,
                 hyper_params: Dict[str, Any] = None,
                 output: str = 'model',
                 spec: str = default_specs.RECOGNITION_SPEC,
                 model: Optional[Union['PathLike', str]] = None,
                 load_hyper_parameters: bool = False,
                 len_train_set: int = -1,
                 legacy_polygons: bool = False):
        """
        A LightningModule encapsulating the unsupervised pretraining setup for
        a text recognition model.

        Setup parameters (load, training_data, evaluation_data, ....) are
        named, model hyperparameters (everything in
        `kraken.lib.default_specs.RECOGNITION_HYPER_PARAMS`) are in in the
        `hyper_params` argument.

        Args:
            hyper_params (dict): Hyperparameter dictionary containing all fields
                                 from
                                 kraken.lib.default_specs.RECOGNITION_PRETRAIN_HYPER_PARAMS
            **kwargs: Setup parameters, i.e. CLI parameters of the train() command.
        """
        super().__init__()
        hyper_params_ = default_specs.RECOGNITION_PRETRAIN_HYPER_PARAMS
        self.legacy_polygons = legacy_polygons

        if model:
            logger.info(f'Loading existing model from {model} ')
            self.nn = vgsl.TorchVGSLModel.load_model(model)

            # apply legacy polygon parameter
            self.nn.use_legacy_polygons = legacy_polygons

            if self.nn.model_type not in [None, 'recognition']:
                raise ValueError(f'Model {model} is of type {self.nn.model_type} while `recognition` is expected.')

            if load_hyper_parameters:
                hp = self.nn.hyper_params
            else:
                hp = {}
            hyper_params_.update(hp)
        else:
            self.nn = None

        if hyper_params:
            hyper_params_.update(hyper_params)
        self.save_hyperparameters(hyper_params_)

        self.model = model
        self.output = output
        self.len_train_set = len_train_set

        self.best_epoch = 0
        self.best_metric = math.inf
        self.best_model = None

        self.val_ce = []

        spec = spec.strip()
        if spec[0] != '[' or spec[-1] != ']':
            raise ValueError(f'VGSL spec {spec} not bracketed')
        self.spec = spec
        # preparse input sizes from vgsl string to seed ground truth data set
        # sizes and dimension ordering.
        if not self.nn:
            blocks = spec[1:-1].split(' ')
            m = re.match(r'(\d+),(\d+),(\d+),(\d+)', blocks[0])
            if not m:
                raise ValueError(f'Invalid input spec {blocks[0]}')
            self.batch, self.height, self.width, self.channels = [int(x) for x in m.groups()]
        else:
            self.batch, self.channels, self.height, self.width = self.nn.input

        if 'file_system' in torch.multiprocessing.get_all_sharing_strategies():
            logger.debug('Setting multiprocessing tensor sharing strategy to file_system')
            torch.multiprocessing.set_sharing_strategy('file_system')

        logger.info('Encoding training set')

    def forward(self, x, seq_lens):
        return self.net(x, seq_lens)

    def _step(self, batch, batch_idx):
        try:
            # sequence batch
            if 'seq_lens' in batch:
                output = self.features(batch['image'], batch['seq_lens'])
            else:
                output = self.features(batch['image'])

            # height should be 1 by now
            if output[0].size(2) != 1:
                raise KrakenInputException('Expected dimension 3 to be 1, actual {}'.format(output[0].size(2)))

            mask_output = self.wav2vec2mask(*output)

            # run contextual encoder, i.e. recurrent layers
            output, seq_lens = self.encoder(mask_output['output'], mask_output['seq_len'])

            # unmasked features in encoder output domain
            y = mask_output['unmasked_samples']
            # negative samples
            negatives = mask_output['negative_samples']
            N, C, H, W = output.shape
            output = output.transpose(1, 3).reshape(-1, W, C)
            # masked features after encoder
            x = output[mask_output['mask']].reshape_as(y)
            mask_n_neg = torch.cat([y.unsqueeze(0), negatives], dim=0)
            logits = torch.cosine_similarity(x.float(), mask_n_neg.float(), dim=-1).type_as(x)

            targets = logits.new_zeros(logits.size(1) * logits.size(2), dtype=torch.long)

            logits = logits.transpose(0, 2)
            logits = logits.reshape(-1, logits.size(-1))
            logits /= self.hparams.logit_temp

            loss = F.cross_entropy(logits, targets)
            return logits, targets, loss
        except RuntimeError as e:
            if is_oom_error(e):
                logger.warning('Out of memory error in trainer. Skipping batch and freeing caches.')
                garbage_collection_cuda()
            else:
                raise

    def validation_step(self, batch, batch_idx):
        o = self._step(batch, batch_idx)
        if o is not None:
            logits, targets, loss = o
            self.val_ce.append(loss.cpu())
            self.log('CE', loss, on_step=True, on_epoch=True)

    def on_validation_epoch_end(self):
        if not self.trainer.sanity_checking:
            ce = np.mean(self.val_ce)

            if ce < self.best_metric:
                logger.debug(f'Updating best metric from {self.best_metric} ({self.best_epoch}) to {ce} ({self.current_epoch})')
                self.best_epoch = self.current_epoch
                self.best_metric = ce
            logger.info(f'validation run: cross_enctropy: {ce}')
            self.log('val_ce', ce, on_step=False, on_epoch=True, prog_bar=True, logger=True)
        self.val_ce.clear()

    def training_step(self, batch, batch_idx):
        o = self._step(batch, batch_idx)
        if o is not None:
            _, _, loss = o
            self.log('CE', loss)
            return loss

    def configure_optimizers(self):
        return _configure_optimizer_and_lr_scheduler(self.hparams,
                                                     chain(self.features.parameters(),
                                                           self.wav2vec2mask.parameters(),
                                                           self.encoder.parameters()),
                                                     len_train_set=self.len_train_set,
                                                     loss_tracking_mode='min')

    def optimizer_step(self, epoch, batch_idx, optimizer, optimizer_closure):
        # update params
        optimizer.step(closure=optimizer_closure)

        # linear warmup between 0 and the initial learning rate `lrate` in `warmup`
        # steps.
        if self.hparams.warmup and self.trainer.global_step < self.hparams.warmup:
            lr_scale = min(1.0, float(self.trainer.global_step + 1) / self.hparams.warmup)
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.lrate

    def lr_scheduler_step(self, scheduler, metric):
        if not self.hparams.warmup or self.trainer.global_step >= self.hparams.warmup:
            # step OneCycleLR each batch if not in warmup phase
            if isinstance(scheduler, lr_scheduler.OneCycleLR):
                scheduler.step()
            # step every other scheduler epoch-wise
            elif self.trainer.is_last_batch:
                scheduler.step()

    def setup(self, stage: Optional[str] = None):
        # finalize models in case of appending/loading
        if stage in [None, 'fit']:
            if self.model:
                self.spec = self.nn.spec
            else:
                logger.info(f'Creating new model {self.spec}')
                self.nn = vgsl.TorchVGSLModel(self.spec)
                self.nn.use_legacy_polygons = self.legacy_polygons
                # initialize weights
                self.nn.init_weights()

            self.net = self.nn.nn

            for idx, layer in enumerate(self.net.children()):
                if isinstance(layer, layers.TransposedSummarizingRNN):
                    break

            self.features = self.net[:idx]
            if self.model and 'wav2vec2mask' in self.nn.aux_layers:
                logger.info('Extracting wav2vec2mask layer from model: mask width '
                            f'{self.nn.aux_layers["wav2vec2mask"].mask_width}, prob '
                            f'{self.nn.aux_layers["wav2vec2mask"].mask_prob}, negative samples '
                            f'{self.nn.aux_layers["wav2vec2mask"].num_negatives}')
                self.wav2vec2mask = self.nn.aux_layers['wav2vec2mask']
                logger.info("Overriding masking hyperparameters with model one's: ")
                self.hparams.mask_width = self.wav2vec2mask.mask_width
                self.hparams.mask_mask_prob = self.wav2vec2mask.mask_prob
                self.hparams.num_negatives = self.wav2vec2mask.num_negatives
            else:
                logger.info(f'Instantiating new wav2vec2mask layer: mask width '
                            f'{self.hparams.mask_width}, prob '
                            f'{self.hparams.mask_prob}, negative samples '
                            f'{self.hparams.num_negatives}')
                self.wav2vec2mask = Wav2Vec2Mask(self.net[idx-1].output_shape[1],
                                                 self.net[-1].output_shape[1],
                                                 self.hparams.mask_width,
                                                 self.hparams.mask_prob,
                                                 self.hparams.num_negatives)
                self.nn.aux_layers = {'wav2vec2mask': self.wav2vec2mask}

            # add dummy codec and output layer
            if not self.nn.codec and not isinstance(self.net[-1], layers.LinSoftmax):
                logger.info('Adding dummy codec and output layer to model')
                self.nn.add_codec(PytorchCodec(' '))
                self.nn.append(len(self.net), "[O1c2]")
            self.encoder = self.net[idx:]
            self.nn.hyper_params = self.hparams
            self.nn.model_type = 'recognition'

    def configure_callbacks(self):
        callbacks = []
        if self.hparams.quit == 'early':
            callbacks.append(EarlyStopping(monitor='CE',
                                           mode='min',
                                           patience=self.hparams.lag,
                                           stopping_threshold=0.0))
        return callbacks



#!/usr/bin/env python
"""
Draws transparent character bounding boxes over images giving a legacy
segmenter model.
"""

import os
import sys
from itertools import cycle

from PIL import Image, ImageDraw

from kraken.binarization import nlbin
from kraken.lib import models
from kraken.pageseg import segment
from kraken.rpred import rpred

cmap = cycle([(230, 25, 75, 127),
              (60, 180, 75, 127),
              (255, 225, 25, 127),
              (0, 130, 200, 127),
              (245, 130, 48, 127),
              (145, 30, 180, 127),
              (70, 240, 240, 127)])

net = models.load_any(sys.argv[1])

for fname in sys.argv[2:]:
    im = Image.open(fname)
    print(fname)
    im = nlbin(im)
    res = segment(im, maxcolseps=0)
    pred = rpred(net, im, res)
    im = im.convert('RGBA')
    tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
    draw = ImageDraw.Draw(tmp)
    for line in pred:
        for box in line.cuts:
            draw.rectangle(box, fill=next(cmap))
    im = Image.alpha_composite(im, tmp)
    im.save('high_{}'.format(os.path.basename(fname)))



# -*- coding: utf-8 -*-

import json
import unittest
import tempfile
import pyarrow as pa

from pathlib import Path
from pytest import raises

import kraken
from kraken.lib import xml
from kraken.lib.arrow_dataset import build_binary_dataset

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

def _validate_ds(self, path, num_lines, num_empty_lines, ds_type):
    with pa.memory_map(path, 'rb') as source:
        ds_table = pa.ipc.open_file(source).read_all()
        raw_metadata = ds_table.schema.metadata
        if not raw_metadata or b'lines' not in raw_metadata:
            raise ValueError(f'{file} does not contain a valid metadata record.')
        metadata = json.loads(raw_metadata[b'lines'])
    self.assertEqual(metadata['type'],
                    ds_type,
                    f'Unexpected dataset type (expected: {ds_type}, found: {metadata["type"]}')
    self.assertEqual(metadata['counts']['all'],
                     num_lines,
                     'Unexpected number of lines in dataset metadata '
                     f'(expected: {num_lines}, found: {metadata["counts"]["all"]}')
    self.assertEqual(len(ds_table),
                     num_lines,
                     'Unexpected number of rows in arrow table '
                     f'(expected: {num_lines}, found: {metadata["counts"]["all"]}')

    real_empty_lines = len([line for line in ds_table.column('lines') if not str(line[0])])
    self.assertEqual(real_empty_lines,
                     num_empty_lines,
                     'Unexpected number of empty lines in dataset '
                     f'(expected: {num_empty_lines}, found: {real_empty_lines}')


class TestKrakenArrowCompilation(unittest.TestCase):
    """
    Tests for binary datasets
    """
    def setUp(self):
        self.xml = resources / '170025120000003,0074-lite.xml'
        self.seg = xml.XMLPage(self.xml).to_container()
        self.box_lines = [resources / '000236.png']

    def test_build_path_dataset(self):
        with tempfile.NamedTemporaryFile() as tmp_file:
            build_binary_dataset(files=4*self.box_lines,
                                 output_file=tmp_file.name,
                                 format_type='path')
            _validate_ds(self, tmp_file.name, 4, 0, 'kraken_recognition_bbox')

    def test_build_xml_dataset(self):
        with tempfile.NamedTemporaryFile() as tmp_file:
            build_binary_dataset(files=[self.xml],
                                 output_file=tmp_file.name,
                                 format_type='xml')
            _validate_ds(self, tmp_file.name, 4, 0, 'kraken_recognition_baseline')

    def test_build_seg_dataset(self):
        with tempfile.NamedTemporaryFile() as tmp_file:
            build_binary_dataset(files=[self.seg],
                                 output_file=tmp_file.name,
                                 format_type=None)
            _validate_ds(self, tmp_file.name, 4, 0, 'kraken_recognition_baseline')

    def test_forced_type_dataset(self):
        with tempfile.NamedTemporaryFile() as tmp_file:
            build_binary_dataset(files=4*self.box_lines,
                                 output_file=tmp_file.name,
                                 format_type='path',
                                 force_type='kraken_recognition_baseline')
            _validate_ds(self, tmp_file.name, 4, 0, 'kraken_recognition_baseline')

    def test_build_empty_dataset(self):
        """
        Test that empty lines are retained in compiled dataset.
        """
        with tempfile.NamedTemporaryFile() as tmp_file:
            build_binary_dataset(files=[self.xml],
                                 output_file=tmp_file.name,
                                 format_type='xml',
                                 skip_empty_lines=False)
            _validate_ds(self, tmp_file.name, 5, 1, 'kraken_recognition_baseline')



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Utility functions for data loading and training of VGSL networks.
"""
import json
import numbers
from collections import Counter
from functools import partial
from typing import Any, Dict, List, Sequence, Tuple, Union

import importlib_resources
import torch
import torch.nn.functional as F
from torchvision import transforms

from kraken.lib import functional_im_transforms as F_t
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.lineest import CenterNormalizer

__all__ = ['ImageInputTransforms',
           'collate_sequences']

import logging

logger = logging.getLogger(__name__)


class ImageInputTransforms(transforms.Compose):
    def __init__(self,
                 batch: int,
                 height: int,
                 width: int,
                 channels: int,
                 pad: Union[int, Tuple[int, int], Tuple[int, int, int, int]],
                 valid_norm: bool = True,
                 force_binarization: bool = False) -> None:
        """
        Container for image input transforms for recognition and segmentation
        networks.

        Args:
            batch: mini-batch size
            height: height of input image in pixels
            width: width of input image in pixels
            channels: color channels of input
            pad: Amount of padding on horizontal ends of image
            valid_norm: Enables/disables baseline normalization as a valid
                        preprocessing step. If disabled we will fall back to
                        standard scaling.
            force_binarization: Forces binarization of input images using the
                                nlbin algorithm.
        """
        super().__init__(None)

        self._scale: Tuple[int, int] = (height, width)
        self._valid_norm = valid_norm
        self._force_binarization = force_binarization
        self._batch = batch
        self._channels = channels
        self.pad = pad

        self._create_transforms()

    def _create_transforms(self) -> None:
        height = self._scale[0]
        width = self._scale[1]
        self._center_norm = False
        self._mode = 'RGB' if self._channels == 3 else 'L'
        if height == 1 and width == 0 and self._channels > 3:
            perm = (1, 0, 2)
            self._scale = (self._channels, 0)
            self._channels = 1
            if self._valid_norm:
                self._center_norm = True
            self._mode = 'L'
        elif height > 1 and width == 0 and self._channels in (1, 3):
            perm = (0, 1, 2)
            if self._valid_norm and self._channels == 1:
                self._center_norm = True
        elif height == 0 and width > 1 and self._channels in (1, 3):
            perm = (0, 1, 2)
        # fixed height and width image => bicubic scaling of the input image, disable padding
        elif height > 0 and width > 0 and self._channels in (1, 3):
            perm = (0, 1, 2)
            self._pad = 0
        elif height == 0 and width == 0 and self._channels in (1, 3):
            perm = (0, 1, 2)
            self._pad = 0
        else:
            raise KrakenInputException(f'Invalid input spec {self._batch}, {height}, {width}, {self._channels}, {self._pad}.')

        if self._mode != 'L' and self._force_binarization:
            raise KrakenInputException(f'Invalid input spec {self._batch}, {height}, {width}, {self._channels}, {self._pad} in '
                                       'combination with forced binarization.')

        self.transforms = []
        self.transforms.append(transforms.Lambda(partial(F_t.pil_to_mode, mode=self._mode)))

        if self._force_binarization:
            self.transforms.append(transforms.Lambda(F_t.pil_to_bin))
        if self._scale != (0, 0):
            if self._center_norm:
                lnorm = CenterNormalizer(self._scale[0])
                self.transforms.append(transforms.Lambda(partial(F_t.pil_dewarp, lnorm=lnorm)))
                self.transforms.append(transforms.Lambda(partial(F_t.pil_to_mode, mode=self._mode)))
            else:
                self.transforms.append(transforms.Lambda(partial(F_t.pil_fixed_resize, scale=self._scale)))
        if self._pad:
            self.transforms.append(transforms.Pad(self._pad, fill=255))
        self.transforms.append(transforms.ToTensor())
        # invert
        self.transforms.append(transforms.Lambda(F_t.tensor_invert))
        self.transforms.append(transforms.Lambda(partial(F_t.tensor_permute, perm=perm)))

    @property
    def batch(self) -> int:
        """
        Batch size attribute. Ignored.
        """
        return self._batch

    @batch.setter
    def batch(self, batch: int) -> None:
        self._batch = batch

    @property
    def channels(self) -> int:
        """
        Channels attribute. Can be either 1 (binary/grayscale), 3 (RGB).
        """
        if self._channels not in [1, 3] and self._scale[0] == self._channels:
            return 1
        else:
            return self._channels

    @channels.setter
    def channels(self, channels: int) -> None:
        self._channels = channels
        self._create_transforms()

    @property
    def height(self) -> int:
        """
        Desired output image height. If set to 0, image will be rescaled
        proportionally with width, if 1 and `channels` is larger than 3 output
        will be grayscale and of the height set with the channels attribute.
        """
        if self._scale == (1, 0) and self.channels > 3:
            return self._channels
        else:
            return self._scale[0]

    @height.setter
    def height(self, height: int) -> None:
        self._scale = (height, self.scale[1])
        self._create_transforms()

    @property
    def width(self) -> int:
        """
        Desired output image width. If set to 0, image will be rescaled
        proportionally with height.
        """
        return self._scale[1]

    @width.setter
    def width(self, width: int) -> None:
        self._scale = (self._scale[0], width)
        self._create_transforms()

    @property
    def mode(self) -> str:
        """
        Imaginary PIL.Image.Image mode of the output tensor. Possible values
        are RGB, L, and 1.
        """
        return self._mode if not self.force_binarization else '1'

    @property
    def scale(self) -> Tuple[int, int]:
        """
        Desired output shape (height, width) of the image. If any value is set
        to 0, image will be rescaled proportionally with height, width, if 1
        and `channels` is larger than 3 output will be grayscale and of the
        height set with the channels attribute.
        """
        if self._scale == (1, 0) and self.channels > 3:
            return (self._channels, self._scale[1])
        else:
            return self._scale

    @scale.setter
    def scale(self, scale: Tuple[int, int]) -> None:
        self._scale = scale
        self._create_transforms()

    @property
    def pad(self) -> int:
        """
        Amount of padding around left/right end of image.
        """
        return self._pad

    @pad.setter
    def pad(self, pad: Union[int, Tuple[int, int], Tuple[int, int, int, int]]) -> None:
        if not isinstance(pad, (numbers.Number, tuple, list)):
            raise TypeError('Got inappropriate padding arg')
        self._pad = pad
        self._create_transforms()

    @property
    def valid_norm(self) -> bool:
        """
        Switch allowing/disallowing centerline normalization. Even if enabled
        won't be applied to 3-channel images.
        """
        return self._valid_norm

    @valid_norm.setter
    def valid_norm(self, valid_norm: bool) -> None:
        self._valid_norm = valid_norm
        self._create_transforms()

    @property
    def centerline_norm(self) -> bool:
        """
        Attribute indicating if centerline normalization will be applied to
        input images.
        """
        return self._center_norm

    @property
    def force_binarization(self) -> bool:
        """
        Switch enabling/disabling forced binarization.
        """
        return self._force_binarization

    @force_binarization.setter
    def force_binarization(self, force_binarization: bool) -> None:
        self._force_binarization = force_binarization
        self._create_transforms()


def global_align(seq1: Sequence[Any], seq2: Sequence[Any]) -> Tuple[int, List[str], List[str]]:
    """
    Computes a global alignment of two strings.

    Args:
        seq1 (Sequence[Any]):
        seq2 (Sequence[Any]):

    Returns a tuple (distance, list(algn1), list(algn2))
    """
    # calculate cost and direction matrix
    cost = [[0] * (len(seq2) + 1) for x in range(len(seq1) + 1)]
    for i in range(1, len(cost)):
        cost[i][0] = i
    for i in range(1, len(cost[0])):
        cost[0][i] = i
    direction = [[(0, 0)] * (len(seq2) + 1) for x in range(len(seq1) + 1)]
    direction[0] = [(0, x) for x in range(-1, len(seq2))]
    for i in range(-1, len(direction) - 1):
        direction[i + 1][0] = (i, 0)
    for i in range(1, len(cost)):
        for j in range(1, len(cost[0])):
            delcost = ((i - 1, j), cost[i - 1][j] + 1)
            addcost = ((i, j - 1), cost[i][j - 1] + 1)
            subcost = ((i - 1, j - 1), cost[i - 1][j - 1] + (seq1[i - 1] != seq2[j - 1]))
            best = min(delcost, addcost, subcost, key=lambda x: x[1])
            cost[i][j] = best[1]
            direction[i][j] = best[0]
    d = cost[-1][-1]
    # backtrace
    algn1: List[Any] = []
    algn2: List[Any] = []
    i = len(direction) - 1
    j = len(direction[0]) - 1
    while direction[i][j] != (-1, 0):
        k, m = direction[i][j]
        if k == i - 1 and m == j - 1:
            algn1.insert(0, seq1[i - 1])
            algn2.insert(0, seq2[j - 1])
        elif k < i:
            algn1.insert(0, seq1[i - 1])
            algn2.insert(0, '')
        elif m < j:
            algn1.insert(0, '')
            algn2.insert(0, seq2[j - 1])
        i, j = k, m
    return d, algn1, algn2


def compute_confusions(algn1: Sequence[str], algn2: Sequence[str]):
    """
    Compute confusion matrices from two globally aligned strings.

    Args:
        align1 (Sequence[str]): sequence 1
        align2 (Sequence[str]): sequence 2

    Returns:
        A tuple (counts, scripts, ins, dels, subs) with `counts` being per-character
        confusions, `scripts` per-script counts, `ins` a dict with per script
        insertions, `del` an integer of the number of deletions, `subs` per
        script substitutions.
    """
    counts: Dict[Tuple[str, str], int] = Counter()
    ref = importlib_resources.files(__name__).joinpath('scripts.json')
    with ref.open('rb') as fp:
        script_map = json.load(fp)

    def _get_script(c):
        for s, e, n in script_map:
            if ord(c) == s or (e and s <= ord(c) <= e):
                return n
        return 'Unknown'

    scripts: Dict[Tuple[str, str], int] = Counter()
    ins: Dict[Tuple[str, str], int] = Counter()
    dels: int = 0
    subs: Dict[Tuple[str, str], int] = Counter()
    for u, v in zip(algn1, algn2):
        counts[(u, v)] += 1
    for k, v in counts.items():
        if k[0] == '':
            dels += v
        else:
            script = _get_script(k[0])
            scripts[script] += v
            if k[1] == '':
                ins[script] += v
            elif k[0] != k[1]:
                subs[script] += v
    return counts, scripts, ins, dels, subs


def collate_sequences(batch):
    """
    Sorts and pads sequences.
    """
    sorted_batch = sorted(batch, key=lambda x: x['image'].shape[2], reverse=True)
    seqs = [x['image'] for x in sorted_batch]
    seq_lens = torch.LongTensor([seq.shape[2] for seq in seqs])
    max_len = seqs[0].shape[2]
    seqs = torch.stack([F.pad(seq, pad=(0, max_len-seq.shape[2])) for seq in seqs])
    if isinstance(sorted_batch[0]['target'], str):
        labels = [x['target'] for x in sorted_batch]
    else:
        labels = torch.cat([x['target'] for x in sorted_batch]).long()
    label_lens = torch.LongTensor([len(x['target']) for x in sorted_batch])
    return {'image': seqs, 'target': labels, 'seq_lens': seq_lens, 'target_lens': label_lens}



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from typing import Sequence, Union

import numpy as np
import torch


def positive_integers_with_sum(n, total):
    ls = [0]
    rv = []
    while len(ls) < n:
        c = random.randint(0, total)
        ls.append(c)
    ls = sorted(ls)
    ls.append(total)
    for i in range(1, len(ls)):
        rv.append(ls[i] - ls[i-1])
    return rv


def compute_masks(mask_prob: int,
                  mask_width: int,
                  num_neg_samples: int,
                  seq_lens: Union[torch.Tensor, Sequence[int]]):
    """
    Samples num_mask non-overlapping random masks of length mask_width in
    sequence of length seq_len.

    Args:
        mask_prob: Probability of each individual token being chosen as start
                   of a masked sequence. Overall number of masks num_masks is
                   mask_prob * sum(seq_lens) / mask_width.
        mask_width: width of each mask
        num_neg_samples: Number of samples from unmasked sequence parts (gets
                         multiplied by num_mask)
        seq_lens: sequence lengths

    Returns:
        An index array containing 1 for masked bits, 2 for negative samples,
        the number of masks, and the actual number of negative samples.
    """
    mask_samples = np.zeros(sum(seq_lens))
    num_masks = int(mask_prob * sum(seq_lens.numpy()) // mask_width)
    num_neg_samples = num_masks * num_neg_samples
    num_masks += num_neg_samples

    indices = [x+mask_width for x in positive_integers_with_sum(num_masks, sum(seq_lens)-num_masks*mask_width)]
    start = 0
    mask_slices = []
    for i in indices:
        i_start = random.randint(start, i+start-mask_width)
        mask_slices.append(slice(i_start, i_start+mask_width))
        start += i

    neg_idx = random.sample(range(len(mask_slices)), num_neg_samples)
    neg_slices = [mask_slices.pop(idx) for idx in sorted(neg_idx, reverse=True)]

    mask_samples[np.r_[tuple(mask_slices)]] = 1
    mask_samples[np.r_[tuple(neg_slices)]] = 2

    return mask_samples, num_masks - num_neg_samples, num_neg_samples



#!/usr/bin/env python
"""
Reads in a bunch of ALTO documents and repolygonizes the lines contained with
the kraken polygonizer.
"""
import click


@click.command()
@click.option('-f', '--format-type', type=click.Choice(['alto', 'page']), default='alto',
              help='Sets the input document format. In ALTO and PageXML mode all '
              'data is extracted from xml files containing both baselines, polygons, and a '
              'link to source images.')
@click.option('-tl', '--topline', 'topline', show_default=True, flag_value='topline',
              help='Switch for the baseline location in the scripts. '
                   'Set to topline if the data is annotated with a hanging baseline, as is '
                   'common with Hebrew, Bengali, Devanagari, etc. Set to '
                   ' centerline for scripts annotated with a central line.')
@click.option('-cl', '--centerline', 'topline', flag_value='centerline')
@click.option('-bl', '--baseline', 'topline', flag_value='baseline', default='baseline')
@click.argument('files', nargs=-1)
def cli(format_type, topline, files):
    """
    A small script repolygonizing line boundaries in ALTO or PageXML files.
    """
    if len(files) == 0:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()

    from itertools import groupby
    from os.path import splitext

    from lxml import etree
    from PIL import Image

    from kraken.lib import xml
    from kraken.lib.segmentation import calculate_polygonal_environment

    def _repl_alto(fname, polygons):
        with open(fname, 'rb') as fp:
            doc = etree.parse(fp)
            lines = doc.findall('.//{*}TextLine')
            idx = 0
            for line in lines:
                if line.get('BASELINE') is None:
                    continue
                pol = line.find('./{*}Shape/{*}Polygon')
                if pol is not None:
                    if polygons[idx] is not None:
                        pol.attrib['POINTS'] = ' '.join([str(coord) for pt in polygons[idx] for coord in pt])
                    else:
                        pol.attrib['POINTS'] = ''
                idx += 1
            with open(splitext(fname)[0] + '_rewrite.xml', 'wb') as fp:
                doc.write(fp, encoding='UTF-8', xml_declaration=True)

    def _parse_page_coords(coords):
        points = [x for x in coords.split(' ')]
        points = [int(c) for point in points for c in point.split(',')]
        pts = zip(points[::2], points[1::2])
        return [k for k, g in groupby(pts)]

    def _repl_page(fname, polygons):
        with open(fname, 'rb') as fp:
            doc = etree.parse(fp)
            lines = doc.findall('.//{*}TextLine')
            idx = 0
            for line in lines:
                base = line.find('./{*}Baseline')
                if base is not None and not base.get('points').isspace() and len(base.get('points')):
                    try:
                        _parse_page_coords(base.get('points'))
                    except Exception:
                        continue
                else:
                    continue
                pol = line.find('./{*}Coords')
                if pol is not None:
                    if polygons[idx] is not None:
                        pol.attrib['points'] = ' '.join([','.join([str(x) for x in pt]) for pt in polygons[idx]])
                    else:
                        pol.attrib['points'] = ''
                idx += 1
            with open(splitext(fname)[0] + '_rewrite.xml', 'wb') as fp:
                doc.write(fp, encoding='UTF-8', xml_declaration=True)

    if format_type == 'page':
        repl_fn = _repl_page
    else:
        repl_fn = _repl_alto

    topline = {'topline': True,
               'baseline': False,
               'centerline': None}[topline]

    for doc in files:
        click.echo(f'Processing {doc} ')
        seg = xml.XMLPage(doc).to_container()
        im = Image.open(seg.imagename).convert('L')
        baselines = []
        for x in seg.lines:
            bl = x.baseline if x.baseline is not None else [0, 0]
            baselines.append(bl)
        o = calculate_polygonal_environment(im, baselines, scale=(1800, 0), topline=topline)
        repl_fn(doc, o)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
import unittest
from pathlib import Path

from PIL import Image
from pytest import raises

from kraken.lib import xml
from kraken.lib.dataset import BaselineSet, ImageInputTransforms
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.util import is_bitonal

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

def check_output(self, config, im, output_tensor):
    if config['height'] != 0:
        self.assertEqual(config['height'], output_tensor.shape[1])
    if config['width'] != 0:
        self.assertEqual(config['width'], output_tensor.shape[2])
    if config['force_binarization'] or is_bitonal(im):
        self.assertEqual(len(output_tensor.int().unique()), 2)
    if config['channels'] == 3:
        self.assertEqual(output_tensor.shape[0], 3)

class TestBaselineSet(unittest.TestCase):
    """
    Tests for the BaselineSet segmentation dataset class
    """
    def setUp(self):
        self.doc = xml.XMLPage(resources / '170025120000003,0074.xml').to_container()
        self.transforms = ImageInputTransforms(batch=1,
                                               height=200,
                                               width=100,
                                               channels=1,
                                               pad=0)

    def test_baselineset_simple_xml(self):
        """
        Tests simple BaselineSet instantiation
        """
        ds = BaselineSet(im_transforms=self.transforms)
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 10)
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))

    def test_baselineset_simple_valid_baselines(self):
        """
        Test baseline whitelisting in BaselineSet
        """
        # filter out $pac and $pag baseline classes
        ds = BaselineSet(im_transforms=self.transforms,
                         valid_baselines=['$par', '$tip'])
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 8)
        self.assertEqual(set(ds.class_mapping['baselines'].keys()), set(('$tip', '$par')))
        self.assertNotIn('$pac', ds.class_mapping['baselines'])
        self.assertNotIn('$pag', ds.class_mapping['baselines'])
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))

    def test_baselineset_simple_valid_regions(self):
        """
        Test region whitelisting in BaselineSet
        """
        # filter out $tip and $par regions
        ds = BaselineSet(im_transforms=self.transforms,
                         valid_regions=['$pag', '$pac'])
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 8)
        self.assertEqual(set(ds.class_mapping['regions'].keys()), set(('$pag', '$pac')))
        self.assertNotIn('$par', ds.class_mapping['regions'])
        self.assertNotIn('$tip', ds.class_mapping['regions'])
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))

    def test_baselineset_simple_merge_baselines(self):
        """
        Test baseline merging in BaselineSet
        """
        # merge $par into $tip
        ds = BaselineSet(im_transforms=self.transforms,
                         merge_baselines={'$par': '$tip'})
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 9)
        self.assertEqual(set(ds.class_mapping['baselines'].keys()), set(('$tip', '$pag', '$pac')))
        self.assertEqual(len(ds.targets[0]['baselines']['$tip']), 18)
        self.assertNotIn('$par', ds.class_mapping['baselines'])
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))

    def test_baselineset_merge_after_valid_baselines(self):
        """
        Test that filtering with valid_baselines occurs before merging.
        """
        # merge $par and $pac into $tip but discard $par before
        ds = BaselineSet(im_transforms=self.transforms,
                         valid_baselines=('$tip', '$pac'),
                         merge_baselines={'$par': '$tip', '$pac': '$tip'})
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 7)
        self.assertEqual(set(ds.class_mapping['baselines'].keys()), set(('$tip',)))
        self.assertEqual(len(ds.targets[0]['baselines']['$tip']), 26)
        self.assertNotIn('$par', ds.class_mapping['baselines'])
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))

    def test_baselineset_merge_after_valid_regions(self):
        """
        Test that filtering with valid_regions occurs before merging.
        """
        # merge $par and $pac into $tip but discard $par before
        ds = BaselineSet(im_transforms=self.transforms,
                         valid_regions=('$tip', '$pac'),
                         merge_regions={'$par': '$tip', '$pac': '$tip'})
        ds.add(self.doc)
        ds.add(self.doc)

        sample = ds[0]
        self.assertEqual(len(ds), 2)
        self.assertEqual(ds.num_classes, 7)
        self.assertEqual(set(ds.class_mapping['regions'].keys()), set(('$tip',)))
        self.assertEqual(len(ds.targets[0]['regions']['$tip']), 2)
        self.assertNotIn('$par', ds.class_mapping['regions'])
        self.assertEqual(sample['image'].shape, (1, 200, 100))
        self.assertEqual(sample['target'].shape, (ds.num_classes, 200, 100))


class TestInputTransforms(unittest.TestCase):
    """
    Tests for ImageInputTransforms class
    """

    def setUp(self):
        self.im = Image.open(resources / '000236.png')

        self.simple_inst = {'batch': 1,
                            'height': 48,
                            'width': 0,
                            'channels': 1,
                            'pad': (16, 0),
                            'valid_norm': False,
                            'force_binarization': False}

        self.simple_inst_norm = {'batch': 1,
                                 'height': 48,
                                 'width': 0,
                                 'channels': 1,
                                 'pad': (16, 0),
                                 'valid_norm': True,
                                 'force_binarization': False}

        self.simple_inst_rgb = {'batch': 1,
                                'height': 48,
                                'width': 0,
                                'channels': 3,
                                'pad': (16, 0),
                                'valid_norm': False,
                                'force_binarization': False}

        self.simple_inst_norm_rgb = {'batch': 1,
                                     'height': 48,
                                     'width': 0,
                                     'channels': 3,
                                     'pad': (16, 0),
                                     'valid_norm': True,
                                     'force_binarization': False}

        self.channel_height_inst = {'batch': 1,
                                    'height': 1,
                                    'width': 0,
                                    'channels': 72,
                                    'pad': (16, 0),
                                    'valid_norm': False,
                                    'force_binarization': False}

        self.invalid_channels = {'batch': 1,
                                 'height': 48,
                                 'width': 0,
                                 'channels': 4,
                                 'pad': (16, 0),
                                 'valid_norm': False,
                                 'force_binarization': False}

    def test_imageinputtransforms_simple(self):
        """
        Simple ImageInputTransforms instantiation.
        """
        tf = ImageInputTransforms(**self.simple_inst)
        for k, v in self.simple_inst.items():
            self.assertEqual(getattr(tf, k), v)
        self.assertFalse(tf.centerline_norm)
        check_output(self, self.simple_inst, self.im, tf(self.im))

    def test_imageinputtransforms_simple_rgb(self):
        """
        Simple RGB ImageInputTransforms instantiation.
        """
        tf = ImageInputTransforms(**self.simple_inst_rgb)
        for k, v in self.simple_inst_rgb.items():
            self.assertEqual(getattr(tf, k), v)
        self.assertFalse(tf.centerline_norm)
        check_output(self, self.simple_inst_rgb, self.im, tf(self.im))

    def test_imageinputtransforms_norm_rgb(self):
        """
        RGB ImageInputTransforms instantiation with centerline normalization
        valid (but not enabled).
        """
        tf = ImageInputTransforms(**self.simple_inst_norm_rgb)
        for k, v in self.simple_inst_norm_rgb.items():
            self.assertEqual(getattr(tf, k), v)
        self.assertFalse(tf.centerline_norm)
        check_output(self, self.simple_inst_norm_rgb, self.im, tf(self.im))

    def test_imageinputtransforms_simple_norm(self):
        """
        ImageInputTransforms instantiation with centerline normalization valid.
        """
        tf = ImageInputTransforms(**self.simple_inst_norm)
        for k, v in self.simple_inst_norm.items():
            self.assertEqual(getattr(tf, k), v)
        self.assertTrue(tf.centerline_norm)
        check_output(self, self.simple_inst_norm, self.im, tf(self.im))

    def test_imageinputtransforms_channel_height(self):
        """
        ImageInputTransforms with height in channel dimension
        """
        tf = ImageInputTransforms(**self.channel_height_inst)
        for k, v in self.channel_height_inst.items():
            if k == 'channels':
                self.assertEqual(1, tf.channels)
            elif k == 'height':
                self.assertEqual(self.channel_height_inst['channels'], tf.height)
            else:
                self.assertEqual(getattr(tf, k), v)
        self.assertFalse(tf.centerline_norm)
        self.channel_height_inst['height'] = self.channel_height_inst['channels']
        self.channel_height_inst['channels'] = 1
        check_output(self, self.channel_height_inst, self.im, tf(self.im))

    def test_imageinputtransforms_invalid_channels(self):
        """
        ImageInputTransforms instantiation with invalid number of channels
        """
        with raises(KrakenInputException):
            tf = ImageInputTransforms(**self.invalid_channels)




#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos
~~~~~~~~~~~~~

Command line drivers for training functionality.
"""

import logging

import click
from PIL import Image
from rich.traceback import install

from kraken.lib import log

from .dataset import compile
from .linegen import line_generator
from .pretrain import pretrain
from .recognition import test, train
from .repo import publish
from .ro import roadd, rotrain
from .segmentation import segtest, segtrain
from .transcription import extract, transcription

logging.captureWarnings(True)
logger = logging.getLogger('kraken')
# disable annoying lightning worker seeding log messages
logging.getLogger("lightning.fabric.utilities.seed").setLevel(logging.ERROR)
# install rich traceback handler
install(suppress=[click])

# raise default max image size to 20k * 20k pixels
Image.MAX_IMAGE_PIXELS = 20000 ** 2


@click.group()
@click.version_option()
@click.pass_context
@click.option('-v', '--verbose', default=0, count=True)
@click.option('-s', '--seed', default=None, type=click.INT,
              help='Seed for numpy\'s and torch\'s RNG. Set to a fixed value to '
                   'ensure reproducible random splits of data')
@click.option('-r', '--deterministic/--no-deterministic', default=False,
              help="Enables deterministic training. If no seed is given and enabled the seed will be set to 42.")
def cli(ctx, verbose, seed, deterministic):
    ctx.meta['deterministic'] = False if not deterministic else 'warn'
    if seed:
        from lightning.pytorch import seed_everything
        seed_everything(seed, workers=True)
    elif deterministic:
        from lightning.pytorch import seed_everything
        seed_everything(42, workers=True)

    ctx.meta['verbose'] = verbose
    log.set_logger(logger, level=30 - min(10 * verbose, 20))


cli.add_command(compile)
cli.add_command(pretrain)
cli.add_command(train)
cli.add_command(test)
cli.add_command(segtrain)
cli.add_command(segtest)
cli.add_command(publish)
cli.add_command(rotrain)
cli.add_command(roadd)

# deprecated commands
cli.add_command(line_generator)
cli.add_command(extract)
cli.add_command(transcription)

if __name__ == '__main__':
    cli()



"""
kraken.lib.models
~~~~~~~~~~~~~~~~~

Wrapper around TorchVGSLModel including a variety of forward pass helpers for
sequence classification.
"""
from os.path import abspath, expanduser, expandvars
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
import torch

import kraken.lib.ctc_decoder
import kraken.lib.lineest
from kraken.lib.exceptions import (KrakenInputException,
                                   KrakenInvalidModelException)
from kraken.lib.vgsl import TorchVGSLModel

if TYPE_CHECKING:
    from os import PathLike


__all__ = ['TorchSeqRecognizer', 'load_any']

import logging

logger = logging.getLogger(__name__)


class TorchSeqRecognizer(object):
    """
    A wrapper class around a TorchVGSLModel for text recognition.
    """
    def __init__(self,
                 nn: TorchVGSLModel,
                 decoder=kraken.lib.ctc_decoder.greedy_decoder,
                 train: bool = False,
                 device: str = 'cpu'):
        """
        Constructs a sequence recognizer from a VGSL model and a decoder.

        Args:
            nn: Neural network used for recognition.
            decoder: Decoder function used for mapping softmax activations to
                     labels and positions.
            train: Enables or disables gradient calculation and dropout.
            device: Device to run model on.

        Attributes:
            nn: Neural network used for recognition.
            codec: PytorchCodec extracted from the recognition model.

            decoder: Decoder function used for mapping softmax activations to
                     labels and positions.
            train: Enables or disables gradient calculation and dropout.
            device: Device to run model on.
            one_channel_mode: flag indicating if the model expects binary or
                              grayscale input images.
            seg_type: flag indicating if the model expects baseline- or bounding
                      box-derived text line images.

        Raises:
            ValueError: Is raised when the model type is not a sequence recognizer.
        """
        self.nn = nn
        self.kind = ''
        if train is True:
            self.nn.train()
        elif train is False:
            self.nn.eval()
        self.codec = self.nn.codec
        self.decoder = decoder
        self.train = train
        self.device = device
        if nn.model_type not in [None, 'recognition']:
            raise ValueError(f'Models of type {nn.model_type} are not supported by TorchSeqRecognizer')
        self.one_channel_mode = nn.one_channel_mode
        self.seg_type = nn.seg_type
        if self.device:
            self.nn.to(device)

    def to(self, device):
        """
        Moves model to device and automatically loads input tensors onto it.
        """
        self.device = device
        self.nn.to(device)

    def forward(self, line: torch.Tensor, lens: torch.Tensor = None) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """
        Performs a forward pass on a torch tensor of one or more lines with
        shape (N, C, H, W) and returns a numpy array (N, W, C).

        Args:
            line: NCHW line tensor
            lens: Optional tensor containing sequence lengths if N > 1

        Returns:
            Tuple with (N, W, C) shaped numpy array and final output sequence
            lengths.

        Raises:
            KrakenInputException: Is raised if the channel dimension isn't of
                                  size 1 in the network output.
        """
        if self.device:
            line = line.to(self.device)
        o, olens = self.nn.nn(line, lens)
        if o.size(2) != 1:
            raise KrakenInputException('Expected dimension 3 to be 1, actual {}'.format(o.size()))
        self.outputs = o.detach().squeeze(2).float().cpu().numpy()
        if olens is not None:
            olens = olens.cpu().numpy()
        return self.outputs, olens

    def predict(self, line: torch.Tensor, lens: Optional[torch.Tensor] = None) -> List[List[Tuple[str, int, int, float]]]:
        """
        Performs a forward pass on a torch tensor of a line with shape (N, C, H, W)
        and returns the decoding as a list of tuples (string, start, end,
        confidence).

        Args:
            line: NCHW line tensor
            lens: Optional tensor containing sequence lengths if N > 1

        Returns:
            List of decoded sequences.
        """
        o, olens = self.forward(line, lens)
        dec_seqs = []
        if olens is not None:
            for seq, seq_len in zip(o, olens):
                locs = self.decoder(seq[:, :seq_len])
                dec_seqs.append(self.codec.decode(locs))
        else:
            locs = self.decoder(o[0])
            dec_seqs.append(self.codec.decode(locs))
        return dec_seqs

    def predict_string(self, line: torch.Tensor, lens: Optional[torch.Tensor] = None) -> List[str]:
        """
        Performs a forward pass on a torch tensor of a line with shape (N, C, H, W)
        and returns a string of the results.

        Args:
            line: NCHW line tensor
            lens: Optional tensor containing the sequence lengths of the input batch.
        """
        o, olens = self.forward(line, lens)
        dec_strs = []
        if olens is not None:
            for seq, seq_len in zip(o, olens):
                locs = self.decoder(seq[:, :seq_len])
                dec_strs.append(''.join(x[0] for x in self.codec.decode(locs)))
        else:
            locs = self.decoder(o[0])
            dec_strs.append(''.join(x[0] for x in self.codec.decode(locs)))
        return dec_strs

    def predict_labels(self, line: torch.tensor, lens: torch.Tensor = None) -> List[List[Tuple[int, int, int, float]]]:
        """
        Performs a forward pass on a torch tensor of a line with shape (N, C, H, W)
        and returns a list of tuples (class, start, end, max). Max is the
        maximum value of the softmax layer in the region.
        """
        o, olens = self.forward(line, lens)
        oseqs = []
        if olens is not None:
            for seq, seq_len in zip(o, olens):
                oseqs.append(self.decoder(seq[:, :seq_len]))
        else:
            oseqs.append(self.decoder(o[0]))
        return oseqs


def load_any(fname: Union['PathLike', str],
             train: bool = False,
             device: str = 'cpu') -> TorchSeqRecognizer:
    """
    Loads anything that was, is, and will be a valid ocropus model and
    instantiates a shiny new kraken.lib.lstm.SeqRecognizer from the RNN
    configuration in the file.

    Currently it recognizes the following kinds of models:

        * protobuf models containing VGSL segmentation and recognition
          networks.

    Additionally an attribute 'kind' will be added to the SeqRecognizer
    containing a string representation of the source kind. Current known values
    are:

        * vgsl for VGSL models

    Args:
        fname: Path to the model
        train: Enables gradient calculation and dropout layers in model.
        device: Target device

    Returns:
        A kraken.lib.models.TorchSeqRecognizer object.

    Raises:
        KrakenInvalidModelException: if the model is not loadable by any parser.
    """
    nn = None
    fname = abspath(expandvars(expanduser(fname)))
    logger.info('Loading model from {}'.format(fname))
    try:
        nn = TorchVGSLModel.load_model(str(fname))
    except Exception as e:
        raise KrakenInvalidModelException('File {} not loadable by any parser.'.format(fname)) from e
    seq = TorchSeqRecognizer(nn, train=train, device=device)
    seq.kind = 'vgsl'
    return seq


def validate_hyper_parameters(hyper_params):
    """
    Validate some model's hyper parameters and modify them in place if need be.
    """
    if (hyper_params['quit'] == 'fixed' and hyper_params['completed_epochs'] >= hyper_params['epochs']):
        logger.warning('Maximum epochs reached (might be loaded from given model), starting again from 0.')
        hyper_params['completed_epochs'] = 0



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.transcription
~~~~~~~~~~~~~~~~~~~~~~~~~~

Legacy command line drivers for recognition training data annotation.
"""
import logging
import os
import unicodedata
import uuid
from typing import IO, Any, cast

import click
from bidi.algorithm import get_display

from .util import message

logging.captureWarnings(True)
logger = logging.getLogger('kraken')


@click.command('extract', deprecated=True)
@click.pass_context
@click.option('-b', '--binarize/--no-binarize', show_default=True, default=True,
              help='Binarize color/grayscale images')
@click.option('-u', '--normalization', show_default=True,
              type=click.Choice(['NFD', 'NFKD', 'NFC', 'NFKC']), default=None,
              help='Normalize ground truth')
@click.option('-s', '--normalize-whitespace/--no-normalize-whitespace',
              show_default=True, default=True, help='Normalizes unicode whitespace')
@click.option('-n', '--reorder/--no-reorder', default=False, show_default=True,
              help='Reorder transcribed lines to display order')
@click.option('-r', '--rotate/--no-rotate', default=True, show_default=True,
              help='Skip rotation of vertical lines')
@click.option('-o', '--output', type=click.Path(), default='training', show_default=True,
              help='Output directory')
@click.option('--format',
              default='{idx:06d}',
              show_default=True,
              help='Format for extractor output. valid fields are `src` (source file), `idx` (line number), and `uuid` (v4 uuid)')
@click.argument('transcriptions', nargs=-1, type=click.File(lazy=True))
def extract(ctx, binarize, normalization, normalize_whitespace, reorder,
            rotate, output, format, transcriptions):
    """
    Extracts image-text pairs from a transcription environment created using
    ``ketos transcribe``.
    """
    import base64
    from io import BytesIO

    import regex
    from lxml import etree, html
    from PIL import Image

    from kraken import binarization
    from kraken.lib.progress import KrakenProgressBar

    try:
        os.mkdir(output)
    except Exception:
        pass

    text_transforms = []
    if normalization:
        text_transforms.append(lambda x: unicodedata.normalize(normalization, x))
    if normalize_whitespace:
        text_transforms.append(lambda x: regex.sub(r'\s', ' ', x))
    if reorder:
        text_transforms.append(get_display)

    idx = 0
    manifest = []

    with KrakenProgressBar() as progress:
        read_task = progress.add_task('Reading transcriptions', total=len(transcriptions), visible=True if not ctx.meta['verbose'] else False)

        for fp in transcriptions:
            logger.info('Reading {}'.format(fp.name))
            doc = html.parse(fp)
            etree.strip_tags(doc, etree.Comment)
            td = doc.find(".//meta[@itemprop='text_direction']")
            if td is None:
                td = 'horizontal-lr'
            else:
                td = td.attrib['content']

            im = None
            dest_dict = {'output': output, 'idx': 0, 'src': fp.name, 'uuid': str(uuid.uuid4())}
            for section in doc.xpath('//section'):
                img = section.xpath('.//img')[0].get('src')
                fd = BytesIO(base64.b64decode(img.split(',')[1]))
                im = Image.open(fd)
                if not im:
                    logger.info('Skipping {} because image not found'.format(fp.name))
                    break
                if binarize:
                    im = binarization.nlbin(im)
                for line in section.iter('li'):
                    if line.get('contenteditable') and (not u''.join(line.itertext()).isspace() and u''.join(line.itertext())):
                        dest_dict['idx'] = idx
                        dest_dict['uuid'] = str(uuid.uuid4())
                        logger.debug('Writing line {:06d}'.format(idx))
                        l_img = im.crop([int(x) for x in line.get('data-bbox').split(',')])
                        if rotate and td.startswith('vertical'):
                            im.rotate(90, expand=True)
                        l_img.save(('{output}/' + format + '.png').format(**dest_dict))
                        manifest.append((format + '.png').format(**dest_dict))
                        text = u''.join(line.itertext()).strip()
                        for func in text_transforms:
                            text = func(text)
                        with open(('{output}/' + format + '.gt.txt').format(**dest_dict), 'wb') as t:
                            t.write(text.encode('utf-8'))
                        idx += 1

            progress.update(read_task, advance=1)

    logger.info('Extracted {} lines'.format(idx))
    with open('{}/manifest.txt'.format(output), 'w') as fp:
        fp.write('\n'.join(manifest))


@click.command('transcribe', deprecated=True)
@click.pass_context
@click.option('-d', '--text-direction', default='horizontal-lr',
              type=click.Choice(['horizontal-lr', 'horizontal-rl', 'vertical-lr', 'vertical-rl']),
              help='Sets principal text direction', show_default=True)
@click.option('--scale', default=None, type=click.FLOAT)
@click.option('--bw/--orig', default=True, show_default=True,
              help="Put nonbinarized images in output")
@click.option('-m', '--maxcolseps', default=2, type=click.INT, show_default=True)
@click.option('-b/-w', '--black_colseps/--white_colseps', default=False, show_default=True)
@click.option('-f', '--font', default='',
              help='Font family to use')
@click.option('-fs', '--font-style', default=None,
              help='Font style to use')
@click.option('-p', '--prefill', default=None,
              help='Use given model for prefill mode.')
@click.option('--pad', show_default=True, type=(int, int), default=(0, 0),
              help='Left and right padding around lines')
@click.option('-l', '--lines', type=click.Path(exists=True), show_default=True,
              help='JSON file containing line coordinates')
@click.option('-o', '--output', type=click.File(mode='wb'), default='transcription.html',
              help='Output file', show_default=True)
@click.argument('images', nargs=-1, type=click.File(mode='rb', lazy=True))
def transcription(ctx, text_direction, scale, bw, maxcolseps,
                  black_colseps, font, font_style, prefill, pad, lines, output,
                  images):
    """
    Creates transcription environments for ground truth generation.
    """
    import json

    from PIL import Image

    from kraken import binarization, pageseg, rpred, transcribe
    from kraken.lib import models
    from kraken.lib.progress import KrakenProgressBar

    ti = transcribe.TranscriptionInterface(font, font_style)

    if len(images) > 1 and lines:
        raise click.UsageError('--lines option is incompatible with multiple image files')

    if prefill:
        logger.info('Loading model {}'.format(prefill))
        message('Loading ANN', nl=False)
        prefill = models.load_any(prefill)
        message('\u2713', fg='green')

    with KrakenProgressBar() as progress:
        read_task = progress.add_task('Reading images', total=len(images), visible=True if not ctx.meta['verbose'] else False)
        for fp in images:
            logger.info('Reading {}'.format(fp.name))
            im = Image.open(fp)
            if im.mode not in ['1', 'L', 'P', 'RGB']:
                logger.warning('Input {} is in {} color mode. Converting to RGB'.format(fp.name, im.mode))
                im = im.convert('RGB')
            logger.info('Binarizing page')
            im_bin = binarization.nlbin(im)
            im_bin = im_bin.convert('1')
            logger.info('Segmenting page')
            if not lines:
                res = pageseg.segment(im_bin, text_direction, scale, maxcolseps, black_colseps, pad=pad)
            else:
                with click.open_file(lines, 'r') as fp:
                    try:
                        fp = cast('IO[Any]', fp)
                        res = json.load(fp)
                    except ValueError as e:
                        raise click.UsageError('{} invalid segmentation: {}'.format(lines, str(e)))
            if prefill:
                it = rpred.rpred(prefill, im_bin, res.copy())
                preds = []
                logger.info('Recognizing')
                for pred in it:
                    logger.debug('{}'.format(pred.prediction))
                    preds.append(pred)
                ti.add_page(im, res, records=preds)
            else:
                ti.add_page(im, res)
            fp.close()
            progress.update(read_task, advance=1)

    logger.info('Writing transcription to {}'.format(output.name))
    message('Writing output ', nl=False)
    ti.write(output)
    message('\u2713', fg='green')



#!/usr/bin/env python
"""
Draws a transparent overlay of baseline segmenter output over a list of image
files.
"""
import dataclasses
import os
import re
import unicodedata
from collections import defaultdict
from itertools import cycle

import click

cmap = cycle([(230, 25, 75, 127),
              (60, 180, 75, 127)])

bmap = (0, 130, 200, 255)


def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    value = unicodedata.normalize('NFKD', value)
    value = re.sub(r'[^\w\s-]', '', value).strip().lower()
    value = re.sub(r'[-\s]+', '-', value)
    return value


@click.command()
@click.option('-i', '--model', default=None, show_default=True, type=click.Path(exists=True),
              help='Baseline detection model to use. Overrides format type and expects image files as input.')
@click.option('-d', '--text-direction', default='horizontal-lr',
              show_default=True,
              type=click.Choice(['horizontal-lr', 'horizontal-rl',
                                 'vertical-lr', 'vertical-rl']),
              help='Sets principal text direction')
@click.option('--repolygonize/--no-repolygonize', show_default=True,
              default=False, help='Repolygonizes line data in ALTO/PageXML '
              'files. This ensures that the trained model is compatible with the '
              'segmenter in kraken even if the original image files either do '
              'not contain anything but transcriptions and baseline information '
              'or the polygon data was created using a different method. Will '
              'be ignored in `path` mode. Note, that this option will be slow '
              'and will not scale input images to the same size as the segmenter '
              'does.')
@click.option('-tl', '--topline', 'topline', show_default=True, flag_value='topline',
              help='Switch for the baseline location in the scripts. ')
@click.option('-cl', '--centerline', 'topline', flag_value='centerline')
@click.option('-bl', '--baseline', 'topline', flag_value='baseline', default='baseline')
@click.option('--height-scale', default=1800, show_default=True,
              help='Maximum height of input image in height dimension')
@click.argument('files', nargs=-1)
def cli(model, text_direction, repolygonize, topline, height_scale, files):
    """
    A script producing overlays of lines and regions from either ALTO or
    PageXML files or run a model to do the same.
    """
    if len(files) == 0:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()

    from PIL import Image, ImageDraw

    from kraken import blla
    from kraken.lib import segmentation, vgsl, xml

    loc = {'topline': True,
           'baseline': False,
           'centerline': None}

    topline = loc[topline]

    if model is None:
        for doc in files:
            click.echo(f'Processing {doc} ', nl=False)
            data = xml.XMLPage(doc).to_container()
            if repolygonize:
                im = Image.open(data.imagename).convert('L')
                lines = data.lines
                polygons = segmentation.calculate_polygonal_environment(im,
                                                                        [x.baseline for x in lines],
                                                                        scale=(height_scale, 0),
                                                                        topline=topline)
                data.lines = [dataclasses.replace(orig, boundary=polygon) for orig, polygon in zip(lines, polygons)]
            # reorder lines by type
            lines = defaultdict(list)
            for line in data.lines:
                lines[line.tags['type']].append(line)
            im = Image.open(data.imagename).convert('RGBA')
            for t, ls in lines.items():
                tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
                draw = ImageDraw.Draw(tmp)
                for idx, line in enumerate(ls):
                    c = next(cmap)
                    if line.boundary:
                        draw.polygon([tuple(x) for x in line.boundary], fill=c, outline=c[:3])
                    if line.baseline:
                        draw.line([tuple(x) for x in line.baseline], fill=bmap, width=2, joint='curve')
                    draw.text(line.baseline[0], str(idx), fill=(0, 0, 0, 255))
                base_image = Image.alpha_composite(im, tmp)
                base_image.save(f'high_{os.path.basename(doc)}_lines_{slugify(t)}.png')
            for t, regs in data.regions.items():
                tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
                draw = ImageDraw.Draw(tmp)
                for reg in regs:
                    c = next(cmap)
                    try:
                        draw.polygon(reg.boundary, fill=c, outline=c[:3])
                    except Exception:
                        pass
                base_image = Image.alpha_composite(im, tmp)
                base_image.save(f'high_{os.path.basename(doc)}_regions_{slugify(t)}.png')
            click.secho('\u2713', fg='green')
    else:
        net = vgsl.TorchVGSLModel.load_model(model)
        for doc in files:
            click.echo(f'Processing {doc} ', nl=False)
            im = Image.open(doc)
            res = blla.segment(im, model=net, text_direction=text_direction)
            # reorder lines by type
            lines = defaultdict(list)
            for line in res.lines:
                lines[line.tags['type']].append(line)
            im = im.convert('RGBA')
            for t, ls in lines.items():
                tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
                draw = ImageDraw.Draw(tmp)
                for idx, line in enumerate(ls):
                    c = next(cmap)
                    draw.polygon([tuple(x) for x in line.boundary], fill=c, outline=c[:3])
                    draw.line([tuple(x) for x in line.baseline], fill=bmap, width=2, joint='curve')
                    draw.text(line.baseline[0], str(idx), fill=(0, 0, 0, 255))
                base_image = Image.alpha_composite(im, tmp)
                base_image.save(f'high_{os.path.basename(doc)}_lines_{slugify(t)}.png')
            for t, regs in res.regions.items():
                tmp = Image.new('RGBA', im.size, (0, 0, 0, 0))
                draw = ImageDraw.Draw(tmp)
                for reg in regs:
                    c = next(cmap)
                    try:
                        draw.polygon([tuple(x) for x in reg.boundary], fill=c, outline=c[:3])
                    except Exception:
                        pass

                base_image = Image.alpha_composite(im, tmp)
                base_image.save(f'high_{os.path.basename(doc)}_regions_{slugify(t)}.png')
            click.secho('\u2713', fg='green')


if __name__ == '__main__':
    cli()



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Named functions for all the transforms that were lambdas in the past to
facilitate pickling.
"""
import unicodedata
from pathlib import Path
from typing import (TYPE_CHECKING, Any, Callable, Literal, Optional, Tuple,
                    Union)

import bidi.algorithm as bd
import regex
import torch
from PIL.Image import Resampling

from kraken.lib.lineest import CenterNormalizer, dewarp

if TYPE_CHECKING:
    from os import PathLike

    from PIL import Image


def pil_to_mode(im: 'Image.Image', mode: str) -> 'Image.Image':
    return im.convert(mode)


def pil_to_bin(im: 'Image.Image') -> 'Image.Image':
    from kraken.binarization import nlbin
    return nlbin(im)


def dummy(x: Any) -> Any:
    return x


def pil_dewarp(im: 'Image.Image', lnorm: CenterNormalizer) -> 'Image.Image':
    return dewarp(lnorm, im)


def pil_fixed_resize(im: 'Image.Image', scale: Tuple[int, int]) -> 'Image.Image':
    return _fixed_resize(im, scale, Resampling.LANCZOS)


def tensor_invert(im: torch.Tensor) -> torch.Tensor:
    return im.max() - im


def tensor_permute(im: torch.Tensor, perm: Tuple[int, ...]) -> torch.Tensor:
    return im.permute(*perm)


def _fixed_resize(img: 'Image.Image', size: Tuple[int, int], interpolation: int = Resampling.LANCZOS):
    """
    Doesn't do the annoying runtime scale dimension switching the default
    pytorch transform does.

    Args:
        img: image to resize
        size: Tuple (height, width)
    """
    w, h = img.size
    oh, ow = size
    if oh == 0:
        oh = int(h * ow/w)
    elif ow == 0:
        ow = int(w * oh/h)
    img = img.resize((ow, oh), interpolation)
    return img


def text_normalize(text: str, normalization: Literal['NFD', 'NFC', 'NFKD', 'NFKC']) -> str:
    return unicodedata.normalize(normalization, text)


def text_whitespace_normalize(text: str) -> str:
    return regex.sub(r'\s', ' ', text).strip()


def text_reorder(text: str, base_dir: Optional[Literal['L', 'R']] = None) -> str:
    return bd.get_display(text, base_dir=base_dir)


def default_split(x: Union['PathLike', str]) -> str:
    x = Path(x)
    while x.suffixes:
        x = x.with_suffix('')
    return str(x)


def suffix_split(x: Union['PathLike', str], split: Callable[[Union['PathLike', str]], str], suffix: str) -> str:
    return split(x) + suffix



#! /usr/bin/env python
"""
Produces semi-transparent neural segmenter output overlays
"""
import click


@click.command()
@click.option('-i', '--model', default=None, show_default=True, type=click.Path(exists=True),
              help='Baseline detection model to use.')
@click.argument('files', nargs=-1)
def cli(model, files):
    """
    Applies a BLLA baseline segmentation model and outputs the raw heatmaps of
    the first baseline class.
    """
    from os.path import splitext

    import torch
    import torch.nn.functional as F
    import torchvision.transforms as tf
    from PIL import Image

    from kraken.lib import dataset, vgsl

    model = vgsl.TorchVGSLModel.load_model(model)
    model.eval()
    batch, channels, height, width = model.input

    transforms = dataset.ImageInputTransforms(batch, height, width, channels, 0, valid_norm=False)

    torch.set_num_threads(1)

    for img in files:
        print(img)
        im = Image.open(img)
        xs = transforms(im)

        with torch.no_grad():
            o, _ = model.nn(xs.unsqueeze(0))
            o = F.interpolate(o, size=xs.shape[1:])
            o = o.squeeze().numpy()

        scal_im = tf.ToPILImage()(1-xs)
        heat = Image.fromarray((o[2]*255).astype('uint8'))
        heat.save(splitext(img)[0] + '.heat.png')
        overlay = Image.new('RGBA', scal_im.size, (0, 130, 200, 255))
        bl = Image.composite(overlay, scal_im.convert('RGBA'), heat)
        heat = Image.fromarray((o[1]*255).astype('uint8'))
        overlay = Image.new('RGBA', scal_im.size, (230, 25, 75, 255))
        bl = Image.composite(overlay, bl, heat)
        heat = Image.fromarray((o[0]*255).astype('uint8'))
        overlay = Image.new('RGBA', scal_im.size, (60, 180, 75, 255))
        Image.composite(overlay, bl, heat).save(splitext(img)[0] + '.overlay.png')
        del o
        del im


if __name__ == '__main__':
    cli()



# Copyright Benjamin Kiessling
# Copyright The PyTorch Lightning team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Handlers for rich-based progress bars.
"""
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

from lightning.pytorch.callbacks.progress.rich_progress import (
    CustomProgress, MetricsTextColumn, RichProgressBar)
from rich import get_console, reconfigure
from rich.default_styles import DEFAULT_STYLES
from rich.progress import (BarColumn, DownloadColumn, Progress, ProgressColumn,
                           TextColumn, TimeElapsedColumn, TimeRemainingColumn)
from rich.text import Text

if TYPE_CHECKING:
    from rich.console import RenderableType
    from rich.style import Style

__all__ = ['KrakenProgressBar', 'KrakenDownloadProgressBar', 'KrakenTrainProgressBar']


class BatchesProcessedColumn(ProgressColumn):
    def __init__(self):
        super().__init__()

    def render(self, task) -> 'RenderableType':
        total = task.total if task.total != float("inf") else "--"
        return Text(f"{int(task.completed)}/{total}", style='magenta')


class EarlyStoppingColumn(ProgressColumn):
    """
    A column containing text.
    """

    def __init__(self, trainer):
        self._trainer = trainer
        super().__init__()

    def render(self, task) -> Text:

        text = f'early_stopping: ' \
               f'{self._trainer.early_stopping_callback.wait_count}/{self._trainer.early_stopping_callback.patience} ' \
               f'{self._trainer.early_stopping_callback.best_score:.5f}'
        return Text(text, justify="left")


class KrakenProgressBar(Progress):
    """
    Adaptation of the default rich progress bar to fit with kraken/ketos output.
    """
    def __init__(self, *args, **kwargs):
        columns = [TextColumn("[progress.description]{task.description}"),
                   BarColumn(),
                   TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                   BatchesProcessedColumn(),
                   TimeRemainingColumn(),
                   TimeElapsedColumn()]
        kwargs['refresh_per_second'] = 1
        super().__init__(*columns, *args, **kwargs)


class KrakenDownloadProgressBar(Progress):
    """
    Adaptation of the default rich progress bar to fit with kraken/ketos download output.
    """
    def __init__(self, *args, **kwargs):
        columns = [TextColumn("[progress.description]{task.description}"),
                   BarColumn(),
                   TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
                   DownloadColumn(),
                   TimeRemainingColumn(),
                   TimeElapsedColumn()]
        kwargs['refresh_per_second'] = 1
        super().__init__(*columns, *args, **kwargs)


class KrakenTrainProgressBar(RichProgressBar):
    """
    Adaptation of the default ptl rich progress bar to fit with kraken (segtrain, train) output.

    Args:
        refresh_rate: Determines at which rate (in number of batches) the progress bars get updated.
            Set it to ``0`` to disable the display.
        leave: Leaves the finished progress bar in the terminal at the end of the epoch. Default: False
        console_kwargs: Args for constructing a `Console`
    """
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs, theme=RichProgressBarTheme())

    def _init_progress(self, trainer):
        if self.is_enabled and (self.progress is None or self._progress_stopped):
            self._reset_progress_bar_ids()
            reconfigure(**self._console_kwargs)
            self._console = get_console()
            self._console.clear_live()
            self._metric_component = MetricsTextColumn(trainer,
                                                       self.theme.metrics,
                                                       self.theme.metrics_text_delimiter,
                                                       self.theme.metrics_format)
            columns = self.configure_columns(trainer)
            columns.append(self._metric_component)

            if trainer.early_stopping_callback:
                self._early_stopping_component = EarlyStoppingColumn(trainer)
                columns.append(self._early_stopping_component)

            self.progress = CustomProgress(
                *columns,
                auto_refresh=False,
                disable=self.is_disabled,
                console=self._console,
            )
            self.progress.start()
            # progress has started
            self._progress_stopped = False

    def _get_train_description(self, current_epoch: int) -> str:
        return f"stage {current_epoch}/" \
               f"{self.trainer.max_epochs if self.trainer.model.hparams.hyper_params['quit'] == 'fixed' else '∞'}"


@dataclass
class RichProgressBarTheme:
    """Styles to associate to different base components.

    Args:
        description: Style for the progress bar description. For eg., Epoch x, Testing, etc.
        progress_bar: Style for the bar in progress.
        progress_bar_finished: Style for the finished progress bar.
        progress_bar_pulse: Style for the progress bar when `IterableDataset` is being processed.
        batch_progress: Style for the progress tracker (i.e 10/50 batches completed).
        time: Style for the processed time and estimate time remaining.
        processing_speed: Style for the speed of the batches being processed.
        metrics: Style for the metrics

    https://rich.readthedocs.io/en/stable/style.html
    """

    description: Union[str, 'Style'] = DEFAULT_STYLES['progress.description']
    progress_bar: Union[str, 'Style'] = DEFAULT_STYLES['bar.complete']
    progress_bar_finished: Union[str, 'Style'] = DEFAULT_STYLES['bar.finished']
    progress_bar_pulse: Union[str, 'Style'] = DEFAULT_STYLES['bar.pulse']
    batch_progress: Union[str, 'Style'] = DEFAULT_STYLES['progress.description']
    time: Union[str, 'Style'] = DEFAULT_STYLES['progress.elapsed']
    processing_speed: Union[str, 'Style'] = DEFAULT_STYLES['progress.data.speed']
    metrics: Union[str, 'Style'] = DEFAULT_STYLES['progress.description']
    metrics_text_delimiter: str = ' '
    metrics_format: str = '.3f'



"""
Ocropus's magic PIL-numpy array conversion routines. They express slightly
different behavior from PIL.Image.toarray().
"""
import unicodedata
import uuid
from typing import TYPE_CHECKING, Callable, Literal, Optional, Union

import numpy as np
import torch
from PIL import Image

from kraken.containers import BBoxLine
from kraken.lib import functional_im_transforms as F_t
from kraken.lib.exceptions import KrakenInputException

if TYPE_CHECKING:
    from os import PathLike

__all__ = ['pil2array', 'array2pil', 'is_bitonal', 'make_printable', 'get_im_str', 'parse_gt_path']


def pil2array(im: Image.Image, alpha: int = 0) -> np.ndarray:
    if im.mode == '1':
        return np.array(im.convert('L'))
    return np.array(im)


def array2pil(a: np.ndarray) -> Image.Image:
    if a.dtype == np.dtype("B"):
        if a.ndim == 2:
            return Image.frombytes("L", (a.shape[1], a.shape[0]),
                                   a.tobytes())
        elif a.ndim == 3:
            return Image.frombytes("RGB", (a.shape[1], a.shape[0]),
                                   a.tobytes())
        else:
            raise Exception("bad image rank")
    elif a.dtype == np.dtype('float32'):
        return Image.frombytes("F", (a.shape[1], a.shape[0]), a.tobytes())
    else:
        raise Exception("unknown image type")


def is_bitonal(im: Union[Image.Image, torch.Tensor]) -> bool:
    """
    Tests a PIL image or torch tensor for bitonality.

    Args:
        im: Image to test

    Returns:
        True if the image contains only two different color values. False
        otherwise.
    """
    if isinstance(im, Image.Image):
        return im.getcolors(2) is not None and len(im.getcolors(2)) == 2
    elif isinstance(im, torch.Tensor):
        return len(im.unique()) == 2


def get_im_str(im: Image.Image) -> str:
    return im.filename if hasattr(im, 'filename') else str(im)


def is_printable(char: str) -> bool:
    """
    Determines if a chode point is printable/visible when printed.

    Args:
        char (str): Input code point.

    Returns:
        True if printable, False otherwise.
    """
    letters = ('LC', 'Ll', 'Lm', 'Lo', 'Lt', 'Lu')
    numbers = ('Nd', 'Nl', 'No')
    punctuation = ('Pc', 'Pd', 'Pe', 'Pf', 'Pi', 'Po', 'Ps')
    symbol = ('Sc', 'Sk', 'Sm', 'So')
    printable = letters + numbers + punctuation + symbol

    return unicodedata.category(char) in printable


def make_printable(char: str) -> str:
    """
    Takes a Unicode code point and return a printable representation of it.

    Args:
        char (str): Input code point

    Returns:
        Either the original code point, the name of the code point if it is a
        combining mark, whitespace etc., or the hex code if it is a control
        symbol.
    """
    if not char or is_printable(char):
        return char
    elif unicodedata.category(char) in ('Cc', 'Cs', 'Co'):
        return '0x{:x}'.format(ord(char))
    else:
        return unicodedata.name(char)


def parse_gt_path(path: Union[str, 'PathLike'],
                  suffix: str = '.gt.txt',
                  split: Callable[[Union['PathLike', str]], str] = F_t.default_split,
                  skip_empty_lines: bool = True,
                  base_dir: Optional[Literal['L', 'R']] = None,
                  text_direction: Literal['horizontal-lr', 'horizontal-rl', 'vertical-lr', 'vertical-rl'] = 'horizontal-lr') -> BBoxLine:
    """
    Returns a BBoxLine from a image/text file pair.

    Args:
        path: Path to image file
        suffix: Suffix of the corresponding ground truth text file to image
                file in `path`.
        split: Suffix stripping function.
        skip_empty_lines: Whether to raise an exception if ground truth is
                          empty or text file is missing.
        base_dir: Unicode BiDi algorithm base direction
        text_direction: Orientation of the line box.
    """
    try:
        with Image.open(path) as im:
            w, h = im.size
    except Exception as e:
        raise KrakenInputException(e)

    gt = ''
    try:
        with open(F_t.suffix_split(path, split=split, suffix=suffix), 'r', encoding='utf-8') as fp:
            gt = fp.read().strip('\n\r')
    except FileNotFoundError:
        if not skip_empty_lines:
            raise KrakenInputException(f'No text file found for ground truth line {path}.')

    if not gt and skip_empty_lines:
        raise KrakenInputException(f'No text for ground truth line {path}.')

    return BBoxLine(id=uuid.uuid4(),
                    bbox=((0, 0), (w, 0), (w, h), (0, h)),
                    text=gt,
                    base_dir=base_dir,
                    imagename=path,
                    text_direction=text_direction)



# -*- coding: utf-8 -*-
import os
import pickle
import tempfile
import unittest
from pathlib import Path

from pytest import raises

from kraken.lib import models
from kraken.lib.exceptions import KrakenInvalidModelException

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestModels(unittest.TestCase):
    """
    Testing model loading routines
    """

    def setUp(self):
        self.temp = tempfile.NamedTemporaryFile(delete=False)

    def tearDown(self):
        self.temp.close()
        os.unlink(self.temp.name)

    def test_load_invalid(self):
        """
        Tests correct handling of invalid files.
        """
        with raises(KrakenInvalidModelException):
            models.load_any(self.temp.name)
